library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.all;
use work.LZ4CompressorLib.all;
use work.LZ4CompressorOutBufferLib.all;

entity LZ4CompressorOutBufferInControl is
  generic (
    C_BYTEFIFO_ADDR_LENGTH : natural;
    C_BLOCKFIFO_ADDR_LENGTH : natural;
    C_DATA_WIDTH : natural
  );
  port(
    clk : in  std_logic;
    rst : in  std_logic;
    literal_data_valid : in std_logic_vector(C_DATA_WIDTH/8-1 downto 0);
    literal_data_valid_flag : in std_logic;
    literal_length_valid : in std_logic;
    m_proc_ready : out std_logic;
    data_last : in std_logic;
    byte_full : in std_logic;
    last_block : out std_logic;
    header_type : out OutputHeaderType;
    byte_write_ptr : out unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    byte_write_word_ptr : out unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    block_write_ptr : out unsigned(C_BLOCKFIFO_ADDR_LENGTH-1 downto 0);
    last_literal_save : out unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    literal_length : out unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0)
  );
end entity; -- end LZ4CompressorOutBufferInControl

architecture Behavioral of LZ4CompressorOutBufferInControl is
  	type FSM_STATE is (IDLE, FRAME_START, W_IDLE, UNCOMPRESSED_HEADER,
                       SEQUENCE_HEADER, UNFINISHED_HEADER,
                       RESERVE_BLOCKSIZE, FILL_BLOCKSIZE, END_SEQ, DONE);
  	signal current_state : FSM_STATE;
  	signal next_state : FSM_STATE;

    signal block_write_ptr_out : unsigned(C_BLOCKFIFO_ADDR_LENGTH-1 downto 0);
    signal byte_write_ptr_out : unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    signal byte_write_word_ptr_out : unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    signal last_literal_save_out : unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);

    signal block_compressed : boolean;
    signal block_overflow : boolean;
    signal rst_block_cntr : std_logic;

    signal next_block : boolean;
    signal block_counter : natural range 0 to 65535;

    signal finished : boolean;


    signal sequence_counter : unsigned(31 downto 0);
begin
  literal_length <= byte_write_ptr_out - last_literal_save_out;

  BLOCK_CNTR : process(clk)
  begin
    if rising_edge(clk) then
      if rst = '1' or rst_block_cntr = '1' then
        block_counter <= 0;
        block_overflow <= false;
      else
        if literal_data_valid_flag = '1' and byte_full = '0' then
          block_counter <= block_counter+C_DATA_WIDTH/8;
        end if;
        if block_counter >= 65000 then
          block_overflow <= true;
        else
          block_overflow <= false;
        end if;
      end if;
    end if;
  end process; -- end BLOCK_CNTR

  LIT_FIFO : process(clk)
  begin
    if rising_edge(clk) then
      if rst = '1' then
        byte_write_ptr_out <= (others => '0');
        byte_write_word_ptr_out <= (others => '0');
        finished <= false;
      else
        if literal_data_valid_flag = '1' and byte_full = '0' and finished = false then
          byte_write_ptr_out <= byte_write_ptr_out + HighestOne(literal_data_valid);

          if data_last = '1' then
            finished <= true;
          end if;
        end if;
      end if;
    end if;
  end process; -- end LIT_FIFO

  SEQUENCE_COUNTER_PROCESS : process(clk)
  begin
    if rising_edge(clk) then
      if rst = '1' then
        sequence_counter <= (others => '0');
      else
        if literal_length_valid = '1' and current_state = W_IDLE then
          sequence_counter <= sequence_counter + 1;
        end if;
      end if;
    end if;
  end process; -- end SEQUENCE_COUNTER_PROCESS

  FSM_SEQ : process(clk)
	begin
		if rising_edge(clk) then
			if rst = '1' then
				current_state <= IDLE;
        block_write_ptr_out <= (others => '0');
				last_literal_save_out <= (others => '0');
        block_compressed <= false;
        last_block <= '0';
			else
				current_state <= next_state;
        next_block <= false;

        assert literal_length_valid = '0' or current_state = W_IDLE  report "Match ignored" severity warning;

				case(next_state) is
          when SEQUENCE_HEADER =>
            next_block <= true;
						block_write_ptr_out <= block_write_ptr_out +1;
            last_block <= data_last;
					when others =>
				end case;

        case( current_state ) is
          when SEQUENCE_HEADER | UNFINISHED_HEADER |
               UNCOMPRESSED_HEADER =>
               last_literal_save_out <= byte_write_ptr_out;
          when others =>
        end case; -- end case current_state


        case(current_state) is --Need one clock offset for data to be written
          when UNCOMPRESSED_HEADER | FILL_BLOCKSIZE |
               RESERVE_BLOCKSIZE | UNFINISHED_HEADER |
               END_SEQ | FRAME_START =>
            next_block <= true;
            block_write_ptr_out <= block_write_ptr_out +1;
            last_block <= data_last;
          when others =>
        end case;

				case( current_state ) is
					when IDLE =>
						last_literal_save_out <= byte_write_ptr_out;
          when SEQUENCE_HEADER =>
            block_compressed <= true;
          when FILL_BLOCKSIZE =>
            block_compressed <= false;
					when others =>
				end case; -- end case input_current_state
			end if;
		end if;
	end process; -- end INPUT_SEQ

  FSM_COMB_NEXT : process(all)
  begin
  	case( current_state ) is
      when IDLE =>
        next_state <= FRAME_START;
      when FRAME_START =>
        next_state <= RESERVE_BLOCKSIZE;

      when RESERVE_BLOCKSIZE =>
        next_state <= W_IDLE;
      when FILL_BLOCKSIZE =>
        if data_last = '0' then
          next_state <= RESERVE_BLOCKSIZE;
        else
          next_state <= END_SEQ;
        end if;
  		when W_IDLE =>
  			if literal_length_valid = '1' then
  				next_state <= SEQUENCE_HEADER;
        elsif byte_full = '1' or data_last = '1' or block_overflow = true then
          if block_compressed = true then
            next_state <= UNFINISHED_HEADER;
          else
            next_state <= UNCOMPRESSED_HEADER;
          end if;
  			end if;
  		when SEQUENCE_HEADER =>
  			next_state <= W_IDLE;
      when UNFINISHED_HEADER =>
        next_state <= FILL_BLOCKSIZE;
      when UNCOMPRESSED_HEADER =>
        next_state <= FILL_BLOCKSIZE;
      when END_SEQ =>
        next_state <= DONE;
  	  when others =>
  	    next_state <= next_state;
  	end case; -- end case input_current_state
  end process; -- end INPUT_KOMB_NEXT

  FSM_COMB_OUT : process(current_state, next_state)
  begin
    rst_block_cntr <= '0';
    m_proc_ready <= '0';


    case( current_state ) is
      when FRAME_START =>
        header_type <= WriteFrameHeader;
      when SEQUENCE_HEADER =>
        header_type <= WriteMatchSequence;
      when UNFINISHED_HEADER =>
        header_type <= WriteUnfinishedSequence;
      when UNCOMPRESSED_HEADER =>
        header_type <= WriteLiterals;
      when FILL_BLOCKSIZE =>
        header_type <= FillBlockSize;
      when RESERVE_BLOCKSIZE =>
        header_type <= ReserveBlockSize;
      when END_SEQ =>
        header_type <= WriteZeroBlock;
      when others =>
    end case; -- end case next_state

    case( next_state ) is
      when SEQUENCE_HEADER =>
        header_type <= WriteMatchSequence;
      when others =>
    end case; -- end case next_state

    case( current_state ) is
      when IDLE =>
      when FRAME_START =>
      when W_IDLE =>
        m_proc_ready <= '1';
      when SEQUENCE_HEADER =>
      when UNFINISHED_HEADER =>
        rst_block_cntr <= '1';
      when UNCOMPRESSED_HEADER =>
        rst_block_cntr <= '1';
      when others =>
    end case; -- end case input_current_state
  end process; -- end FSM_COMB_OUT

  block_write_ptr <= block_write_ptr_out;
  byte_write_ptr <= byte_write_ptr_out;
  byte_write_word_ptr <= byte_write_word_ptr_out;
  last_literal_save <= last_literal_save_out;
end architecture; -- Behavioral
