library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.LZ4CompressorLib.all;
use work.LZ4CompressorOutBufferLib.all;
use work.all;

entity LZ4CompressorOutBufferOutControl is
  generic (
    C_BYTEFIFO_ADDR_LENGTH : natural;
    C_BLOCKFIFO_ADDR_LENGTH : natural;
    C_DATA_WIDTH : natural;
    C_BLOCK_LENGTH : natural
  );
  port(
    clk : in  std_logic;
    rst : in  std_logic;
    byte_read_ptr : out unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    block_read_ptr : out unsigned(C_BLOCKFIFO_ADDR_LENGTH-1 downto 0);
    block_valid : in std_logic;
    block_type : in std_logic;
    literal_start : in unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    literal_length : in unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    current_block_size : in unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    current_header_type : in OutputHeaderType;
    match_offset : in LZ4Address;
    match_length : in LZ4MatchLength;
    lisc_input_number : out LZ4MatchLength;
    lisc_start : out std_logic;
    lisc_done : in std_logic;
    literals_valid : out std_logic_vector(C_DATA_WIDTH/8-1 downto 0);
    last_block : in std_logic;
		blocks_done : out std_logic;
    write_literal : out std_logic;
    write_lisc : out std_logic;
    write_token : out std_logic;
    write_match : out std_logic;
    write_block_length : out std_logic;
    write_block_reserve : out std_logic;
    block_uncompressed : out std_logic;
    write_done : out std_logic;
    write_frame_header : out std_logic;
    write_end_mark : out std_logic;
    output_buffer_full : in std_logic;
    token_output : out std_logic_vector(7 downto 0);
    output_done : out std_logic
  );
end entity; -- end LZ4CompressorOutBufferOutControl

architecture Behavioral of LZ4CompressorOutBufferOutControl is
  	type FSM_STATE is (IDLE, FRAME_HEADER, TOKEN, LITERAL_LISC_START,
      LITERAL_LISC, LITERALS, MATCH, MATCH_LISC_START, MATCH_LISC, NEXT_HEADER,
      PROCESS_NEXT_HEADER, BLOCK_LENGTH, DONE, RESERVE_BLOCK,
      ENDMARK, FINISHED);
  	signal current_state : FSM_STATE;
  	signal next_state : FSM_STATE;

    signal byte_read_ptr_out : unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    signal block_read_ptr_out : unsigned(C_BLOCKFIFO_ADDR_LENGTH-1 downto 0);
    signal literals_index : unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    signal last_block_uncompressed : boolean;

    signal literals_valid_out : std_logic_vector(C_DATA_WIDTH/8-1 downto 0);

    constant MAX_BLOCK_SIZE : natural := 2**C_BYTEFIFO_ADDR_LENGTH;
    constant BLOCK_TYPE_UNCOMPRESSED : std_logic := '1';
begin
  block_read_ptr <= block_read_ptr_out;
  byte_read_ptr <= byte_read_ptr_out;

  SEQ : process(clk)
  begin
  	if rising_edge(clk) then
  		if rst = '1' then
  			current_state <= IDLE;
  			block_read_ptr_out <= (others => '0');
  			byte_read_ptr_out <= (others => '0');
        last_block_uncompressed <= false;
  		elsif(output_buffer_full = '0') then
        literals_valid <= literals_valid_out;
  			current_state <= next_state;
        write_literal <= '0';

        case( current_state ) is
          when IDLE =>
            literals_index <= (others => '0');
          when TOKEN =>
            byte_read_ptr_out <= literal_start;
          when LITERALS =>
            if next_state = LITERALS then
              write_literal <= '1';
            end if;
            byte_read_ptr_out <= byte_read_ptr_out+C_DATA_WIDTH/8;
            literals_index <= literals_index+C_DATA_WIDTH/8;

            if literals_index+C_DATA_WIDTH/8 >= literal_length or literal_length = 0 then
              --literals_index <= literal_length-1;
              byte_read_ptr_out <= literal_start + literal_length;
            end if;

            if current_header_type = WriteLiterals then
              last_block_uncompressed <= true;
            end if;
          when NEXT_HEADER =>
            byte_read_ptr_out <= literal_start + literal_length;
            block_read_ptr_out <= block_read_ptr_out + 1;
          when MATCH =>
            if (match_length-4) < 15 then
              byte_read_ptr_out <= literal_start + literal_length;
              block_read_ptr_out <= block_read_ptr_out + 1;
            end if;
          when MATCH_LISC =>
            if lisc_done = '1' then
              byte_read_ptr_out <= literal_start + literal_length;
              block_read_ptr_out <= block_read_ptr_out + 1;
            end if;
          when BLOCK_LENGTH =>
            last_block_uncompressed <= false;
          when DONE =>
          when others =>
        end case; -- end case current_state

  		end if;
  	end if;
  end process; -- end SEQ

  NEXT_COMB : process(all)
  begin
    next_state <= current_state;
  	case( current_state ) is
      when IDLE => -- Crossroad for Type of Output to do
        if block_valid = '1' then
          case( current_header_type ) is
            when WriteFrameHeader =>
              next_state <= FRAME_HEADER;
            when ReserveBlockSize =>
              next_state <= RESERVE_BLOCK;
            when WriteMatchSequence | WriteUnfinishedSequence =>
              next_state <= TOKEN;
            when WriteLiterals =>
              next_state <= LITERALS;
            when FillBlockSize =>
              next_state <= BLOCK_LENGTH;
            when WriteZeroBlock =>
              next_state <= ENDMARK;
            when others =>
          end case; -- end case current_header_type
        elsif last_block = '1' then
          next_state <= FINISHED;
        end if;

      -- One way states
      when FRAME_HEADER =>
        next_state <= NEXT_HEADER; --Write Header
      when RESERVE_BLOCK =>
        next_state <= NEXT_HEADER; --Move Write Pointer +4
      when BLOCK_LENGTH =>
        next_state <= NEXT_HEADER; --Write to last valid pointer BlockLength
                                   --Write pointer -> Valid pointer (write_done)
      when ENDMARK =>
        next_state <= NEXT_HEADER; --Write zeros to output stream
                                   --Write pointer -> Valid pointer (write_done)

      when NEXT_HEADER =>
        next_state <= IDLE;

      -- Sequence
      when TOKEN =>
  			if literal_length >= 15 then
  				next_state <= LITERAL_LISC;
  			else
  				next_state <= LITERALS;
  			end if;
  		when LITERAL_LISC =>
  			if lisc_done = '1' then
  				next_state <= LITERALS;
  			end if;
  		when LITERALS =>
  			if literals_index >= literal_length or literal_length = 0 then
          if current_header_type = WriteLiterals or
             current_header_type = WriteUnfinishedSequence then
            next_state <= NEXT_HEADER;
          else
            next_state <= MATCH;
          end if;
  			end if;
      when MATCH =>
        if (match_length-4) >= 15 then
          next_state <= MATCH_LISC;
        else
          next_state <= IDLE;
        end if;
  		when MATCH_LISC =>
  			if lisc_done = '1' then
  				next_state <= IDLE;
  			end if;
  		when others =>
  	end case; -- end case current_state
  end process; -- end NEXT_COMB

  OUT_KOMB : process(all)
    variable tmp_match_length : unsigned(match_length'length-1 downto 0);
    variable tmp_literal_length : unsigned(literal_length'length-1 downto 0);
  begin
    token_output <= (others => '0');
    lisc_input_number <= (others => '0');
    literals_valid_out <= (others => '0');
    write_token <= '0';
    write_lisc <= '0';
    lisc_start <= '0';
    write_match <= '0';
    write_block_length <= '0';
    write_done <= '0';
    write_frame_header <= '0';
    write_end_mark <= '0';
    blocks_done <= '0';
    block_uncompressed <= '0';
    write_block_reserve <= '0';
    output_done <= '0';
    --write_literal <= '0';

    case( next_state ) is
      when others =>
    end case; -- end case next_state
    case( current_state ) is
      when ENDMARK =>
        write_end_mark <= '1';
        write_done <= '1';
      when FINISHED =>
        blocks_done <= '1';
        output_done <= '1';
      when FRAME_HEADER =>
        write_frame_header <= '1';
      when RESERVE_BLOCK =>
        write_block_reserve <= '1';
      when BLOCK_LENGTH =>
        write_block_length <= '1';
        write_done <= '1';
        if(last_block_uncompressed = true) then
          block_uncompressed <= '1';
        end if;

      -- Sequence
      when TOKEN =>
        token_output <= (others => '0');
        if current_header_type = WriteMatchSequence then
          token_output(3 downto 0) <= (others => '1');
          tmp_match_length := (match_length-4);
          if tmp_match_length < 15 then
            token_output(3 downto 0) <= std_logic_vector(tmp_match_length(3 downto 0));
          end if;
        end if;

        token_output(7 downto 4) <= (others => '1');
        if literal_length < 15 then
          token_output(7 downto 4) <= std_logic_vector(literal_length(3 downto 0));
        else
          tmp_literal_length := (literal_length-15);
          lisc_start <= '1';
          lisc_input_number <= resize(tmp_literal_length, lisc_input_number'length);
        end if;
        write_token <= '1';
      when LITERALS =>
        --write_literal <= '1';
        FOR i in 0 to C_DATA_WIDTH/8-1 loop
          if literals_index+i <= literal_length-1 then
            literals_valid_out(i) <= '1';
          end if;
        end loop;
      when MATCH =>
        if(match_length-4 >= 15) then
          tmp_match_length := (match_length-19);
          lisc_start <= '1';
          lisc_input_number <= resize(tmp_match_length, lisc_input_number'length);
        end if;
        write_match <= '1';
      when LITERAL_LISC | MATCH_LISC =>
        write_lisc <= '1';
      when others =>
    end case; -- end case current_state
  end process; -- end OUTPUT_OUT_KOMB
end architecture; -- end Behavioral
