library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;
use work.LZ4CompressorLib.all;
use work.all;

entity LZ4Compressor_v0_1_IN_DATA_AXI is
	generic (
		-- Users to add parameters here
		C_HASHNATOR_COUNT : natural := 8;
		C_HASHNATOR_WIDTH : natural := 32;
		C_HASHNATOR_OFFSET : natural := 8;
		-- User parameters ends
		-- Do not modify the parameters beyond this line

		-- AXI4Stream sink: Data Width
		C_S_AXIS_TDATA_WIDTH	: integer	:= 128
	);
	port (
		-- Users to add ports here
		UNIT_WORKING : in std_logic;
		MEM_MODE : in std_logic;
		MEM_MF_DATA_1 : out std_logic_vector((C_HASHNATOR_WIDTH+(C_HASHNATOR_COUNT-1)*C_HASHNATOR_OFFSET) -1 downto 0);
		MEM_MF_DATA_2 : out std_logic_vector((C_HASHNATOR_WIDTH+(C_HASHNATOR_COUNT-1)*C_HASHNATOR_OFFSET) -1 downto 0);
		MEM_MF_ADDR_1 : in LZ4Address;
		MEM_MF_ADDR_2 : in LZ4Address;

		MEM_ADDR : 			in LZ4Address;
		MEM_HASHING_DATA : out std_logic_vector((C_HASHNATOR_WIDTH+(C_HASHNATOR_COUNT-1)*C_HASHNATOR_OFFSET) -1 downto 0);
		MEM_STREAM_DATA : out std_logic_vector((C_HASHNATOR_WIDTH+(C_HASHNATOR_COUNT-1)*C_HASHNATOR_OFFSET) -1 downto 0);
		MEM_STREAM_DATA_VALID : out std_logic_vector((C_HASHNATOR_WIDTH+(C_HASHNATOR_COUNT-1)*C_HASHNATOR_OFFSET)/8 -1 downto 0);

		MEM_OUT_DATA_VALID : out std_logic_vector((C_HASHNATOR_WIDTH+(C_HASHNATOR_COUNT-1)*C_HASHNATOR_OFFSET)/8 -1 downto 0);
		MEM_OUT_DATA : out std_logic_vector((C_HASHNATOR_WIDTH+(C_HASHNATOR_COUNT-1)*C_HASHNATOR_OFFSET) -1 downto 0);
		MEM_OUT_ADDR : in LZ4Address;
		MEM_LENGTH : out LZ4Address;

		DATA_READY : out std_logic;
		DATA_EMPTY : out std_logic;
		DATA_ENDING : out std_logic;
		-- User ports ends
		-- Do not modify the ports beyond this line

		-- AXI4Stream sink: Clock
		S_AXIS_ACLK	: in std_logic;
		-- AXI4Stream sink: Reset
		S_AXIS_ARESETN	: in std_logic;
		-- Ready to accept data in
		S_AXIS_TREADY	: out std_logic;
		-- Data in
		S_AXIS_TDATA	: in std_logic_vector(C_S_AXIS_TDATA_WIDTH-1 downto 0);
		-- Byte qualifier
		S_AXIS_TSTRB	: in std_logic_vector((C_S_AXIS_TDATA_WIDTH/8)-1 downto 0);
		-- Indicates boundary of last packet
		S_AXIS_TLAST	: in std_logic;
		-- Data is in valid
		S_AXIS_TVALID	: in std_logic
	);
end LZ4Compressor_v0_1_IN_DATA_AXI;

architecture arch_imp of LZ4Compressor_v0_1_IN_DATA_AXI is
	component WordBram
	generic (
	  UNALIGNED_WRITE : boolean := false;
	  PORT_WORD_WIDTH : natural := 8;
	  BYTE_WIDTH      : natural := 8;
	  WORDS_COUNT     : natural := 2048;
	  DEBUG_VIEW      : boolean := true
	);
	port (
	  clk        : in  std_logic;
	  rst        : in  std_logic;
	  addr_in    : in  unsigned(integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT))))-1 downto 0);
	  addr_out   : in  unsigned(integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT)))) -1 downto 0);
	  data_in_we : in  std_logic_vector(PORT_WORD_WIDTH-1 downto 0);
	  data_in    : in  std_logic_vector(PORT_WORD_WIDTH*BYTE_WIDTH-1 downto 0);
	  data_out   : out std_logic_vector(PORT_WORD_WIDTH*BYTE_WIDTH-1 downto 0)
	);
	end component WordBram;

	-- function called clogb2 that returns an integer which has the
	-- value of the ceiling of the log base 2.
	function clogb2 (bit_depth : integer) return integer is
	variable depth  : integer := bit_depth;
	  begin
	    if (depth = 0) then
	      return(0);
	    else
	      for clogb2 in 1 to bit_depth loop  -- Works for up to 32 bit integers
	        if(depth <= 1) then
	          return(clogb2);
	        else
	          depth := depth / 2;
	        end if;
	      end loop;
	    end if;
	end;

	-- Total number of input data.
	constant NUMBER_OF_INPUT_WORDS  : integer := 512;
	constant NUMBER_OF_INPUT_BYTES  : integer := 512*C_S_AXIS_TDATA_WIDTH/8;
	-- bit_num gives the minimum number of bits needed to address 'NUMBER_OF_INPUT_WORDS' size of FIFO.
	constant bit_num  : integer := clogb2(NUMBER_OF_INPUT_BYTES-1);
	-- Define the states of state machine
	-- The control state machine oversees the writing of input streaming data to the FIFO,
	-- and outputs the streaming data from the FIFO
	type state is ( IDLE,        -- This is the initial/idle state
	                WRITE_FIFO); -- In this state FIFO is written with the
	                             -- input stream data S_AXIS_TDATA
	signal axis_tready	: std_logic;
	-- State variable
	signal  mst_exec_state : state;
	-- FIFO implementation signals
	signal  byte_index : integer;
	-- FIFO write enable
	signal fifo_wren : std_logic;
	-- FIFO full flag
	signal fifo_full_flag : std_logic;
	-- FIFO write pointer
	signal write_pointer : natural range 0 to bit_num-1 ;
	-- sink has accepted all the streaming data and stored in FIFO
	signal writes_done : std_logic;

begin
	-- I/O Connections assignments

	S_AXIS_TREADY	<= axis_tready and not UNIT_WORKING;

	-- Control state machine implementation
	process(S_AXIS_ACLK)
	begin
	  if (rising_edge (S_AXIS_ACLK)) then
	    if(S_AXIS_ARESETN = '0') then
	      -- Synchronous reset (active low)
	      mst_exec_state      <= IDLE;
	    else
	      case (mst_exec_state) is
	        when IDLE     =>
	          -- The sink starts accepting tdata when
	          -- there tvalid is asserted to mark the
	          -- presence of valid streaming data
	          if (S_AXIS_TVALID = '1')then
	            mst_exec_state <= WRITE_FIFO;
	          else
	            mst_exec_state <= IDLE;
	          end if;

	        when WRITE_FIFO =>
	          -- When the sink has accepted all the streaming input data,
	          -- the interface swiches functionality to a streaming master
	          if (writes_done = '1') then
	            mst_exec_state <= IDLE;
	          else
	            -- The sink accepts and stores tdata
	            -- into FIFO
	            mst_exec_state <= WRITE_FIFO;
	          end if;

	        when others    =>
	          mst_exec_state <= IDLE;
	      end case;
	    end if;
	  end if;
	end process;
	-- AXI Streaming Sink
	--
	-- The example design sink is always ready to accept the S_AXIS_TDATA  until
	-- the FIFO is not filled with NUMBER_OF_INPUT_WORDS number of input words.
	axis_tready <= '1' when ((mst_exec_state = WRITE_FIFO) and (write_pointer <= NUMBER_OF_INPUT_BYTES-1)) else '0';

	process(S_AXIS_ACLK)
	begin
	  if (rising_edge (S_AXIS_ACLK)) then
	    if(S_AXIS_ARESETN = '0') then
	      write_pointer <= 0;
	      writes_done <= '0';
	    else
	      if (write_pointer <= NUMBER_OF_INPUT_BYTES-(S_AXIS_TSTRB'length*3)) then
	        if (fifo_wren = '1') then
	          -- write pointer is incremented after every write to the FIFO
	          -- when FIFO write signal is enabled.
						for i in 0 to S_AXIS_TSTRB'length-1 loop
							if S_AXIS_TSTRB(i) = '1' then
	          		write_pointer <= write_pointer + i + 1;
							end if;
						end loop;
	          writes_done <= '0';
	        end if;
	      end if;

				if ((write_pointer >= NUMBER_OF_INPUT_BYTES-(S_AXIS_TSTRB'length*4)) or S_AXIS_TLAST = '1') then
					-- reads_done is asserted when NUMBER_OF_INPUT_BYTES numbers of streaming data
					-- has been written to the FIFO which is also marked by S_AXIS_TLAST(kept for optional usage).
					writes_done <= '1';
				end if;
	    end if;
	  end if;
	end process;
	MEM_LENGTH <= to_unsigned(write_pointer, MEM_LENGTH'length);

	-- FIFO write enable generation
	fifo_wren <= S_AXIS_TVALID and axis_tready;


	FIFO_READ : block
		constant bytesInMem : natural := C_S_AXIS_TDATA_WIDTH/C_HASHNATOR_OFFSET;
		constant bytesInStream : natural := (C_HASHNATOR_WIDTH+(C_HASHNATOR_COUNT-1)*C_HASHNATOR_OFFSET)/8;
		signal FIFO_OUT_1 : std_logic_vector((bytesInMem*C_HASHNATOR_OFFSET)-1 downto 0);
		signal FIFO_OUT_2 : std_logic_vector((bytesInMem*C_HASHNATOR_OFFSET)-1 downto 0);
		signal FIFO_OUT_3 : std_logic_vector((bytesInMem*C_HASHNATOR_OFFSET)-1 downto 0);
		signal FIFO_OUT_4 : std_logic_vector((bytesInMem*C_HASHNATOR_OFFSET)-1 downto 0);
		signal FIFO_ADDR_1 : LZ4Address;
		signal FIFO_ADDR_2 : LZ4Address;
		signal FIFO_ADDR_3 : LZ4Address;
		signal FIFO_ADDR_4 : LZ4Address;
		signal FIFO_IN_ADDR : LZ4Address;

		constant BYTE_WIDTH : natural := bytesInMem;
		constant ADDR_IN_WIDTH : natural := integer(ceil(log2(real(BYTE_WIDTH*NUMBER_OF_INPUT_WORDS))));
		constant ADDR_OUT_WIDTH : natural := integer(ceil(log2(real(BYTE_WIDTH*NUMBER_OF_INPUT_WORDS))));
		signal rst : std_logic;
	begin
		rst <= not S_AXIS_ARESETN;
		FIFO_IN_ADDR <= to_unsigned(write_pointer/BYTE_WIDTH, FIFO_IN_ADDR'length);

		WordBram_1 : WordBram
		generic map (
		  PORT_WORD_WIDTH  => BYTE_WIDTH,
		  WORDS_COUNT => NUMBER_OF_INPUT_WORDS
		)
		port map (
		  clk        => S_AXIS_ACLK,
		  rst        => rst,
		  addr_in    => FIFO_IN_ADDR(ADDR_IN_WIDTH-1 downto 0),
		  addr_out   => FIFO_ADDR_1(ADDR_OUT_WIDTH-1 downto 0),
		  data_in_we => S_AXIS_TSTRB,
		  data_in    => S_AXIS_TDATA,
		  data_out   => FIFO_OUT_1
		);

		WordBram_2 : WordBram
		generic map (
		  PORT_WORD_WIDTH  => BYTE_WIDTH,
		  WORDS_COUNT => NUMBER_OF_INPUT_WORDS
		)
		port map (
		  clk        => S_AXIS_ACLK,
		  rst        => rst,
		  addr_in    => FIFO_IN_ADDR(ADDR_IN_WIDTH-1 downto 0),
		  addr_out   => FIFO_ADDR_2(ADDR_OUT_WIDTH-1 downto 0),
		  data_in_we => S_AXIS_TSTRB,
		  data_in    => S_AXIS_TDATA,
		  data_out   => FIFO_OUT_2
		);

		WordBram_3 : WordBram
		generic map (
		  PORT_WORD_WIDTH  => BYTE_WIDTH,
		  WORDS_COUNT => NUMBER_OF_INPUT_WORDS
		)
		port map (
		  clk        => S_AXIS_ACLK,
		  rst        => rst,
		  addr_in    => FIFO_IN_ADDR(ADDR_IN_WIDTH-1 downto 0),
		  addr_out   => FIFO_ADDR_3(ADDR_OUT_WIDTH-1 downto 0),
		  data_in_we => S_AXIS_TSTRB,
		  data_in    => S_AXIS_TDATA,
		  data_out   => FIFO_OUT_3
		);

		WordBram_4 : WordBram
		generic map (
		  PORT_WORD_WIDTH  => BYTE_WIDTH,
		  WORDS_COUNT => NUMBER_OF_INPUT_WORDS
		)
		port map (
		  clk        => S_AXIS_ACLK,
		  rst        => rst,
		  addr_in    => FIFO_IN_ADDR(ADDR_IN_WIDTH-1 downto 0),
		  addr_out   => FIFO_ADDR_4(ADDR_OUT_WIDTH-1 downto 0),
		  data_in_we => S_AXIS_TSTRB,
		  data_in    => S_AXIS_TDATA,
		  data_out   => FIFO_OUT_4
		);

		FIFO_ADDR_1 <= MEM_ADDR;
		FIFO_ADDR_2 <= MEM_MF_ADDR_1;
		FIFO_ADDR_3 <= MEM_MF_ADDR_2;
		FIFO_ADDR_4 <= MEM_OUT_ADDR;

		MEM_STREAM_DATA <= FIFO_OUT_1(MEM_STREAM_DATA'length-1 downto 0);
		MEM_HASHING_DATA <= FIFO_OUT_1(MEM_STREAM_DATA'length-1 downto 0);
	  MEM_MF_DATA_1 <= FIFO_OUT_2(MEM_MF_DATA_1'length-1 downto 0);
	  MEM_MF_DATA_2 <= FIFO_OUT_3(MEM_MF_DATA_2'length-1 downto 0);
	  MEM_OUT_DATA <= FIFO_OUT_4(MEM_OUT_DATA'length-1 downto 0);


		--FIFO_ADDR_MULT : process(MEM_MODE, MEM_ADDR, MEM_MF_ADDR_1, MEM_MF_ADDR_2)
		--begin
		--	if MEM_MODE = '0' then
		--		FIFO_ADDR_1 <= MEM_ADDR;
		--		FIFO_ADDR_2 <= MEM_ADDR;
		--	elsif (MEM_MF_ADDR_1+bytesInStream-1) < NUMBER_OF_INPUT_BYTES then
		--		FIFO_ADDR_1 <= MEM_MF_ADDR_1;
		--		FIFO_ADDR_2 <= MEM_MF_ADDR_2;
		--	else
		--		FIFO_ADDR_1 <= (others => '1');
		--		FIFO_ADDR_2 <= (others => '1');
		--	end if;
		--end process; -- end FIFO_ADDR_MULT


		--FIFO_DATA_MULT : process(MEM_MODE, FIFO_OUT_1, FIFO_OUT_2)
		--begin
		--	MEM_MF_DATA_1 <= (others => '0');
		--	MEM_MF_DATA_2 <= (others => '0');
		--	MEM_STREAM_DATA <= (others => '0');
		--	MEM_HASHING_DATA <= (others => '0');
		--	if MEM_MODE = '0' then
		--		MEM_STREAM_DATA <=  FIFO_OUT_1(MEM_STREAM_DATA'length-1 downto 0);
		--		MEM_HASHING_DATA <= FIFO_OUT_2(MEM_STREAM_DATA'length-1 downto 0);
		--	else
		--		MEM_MF_DATA_1 <= FIFO_OUT_1(MEM_MF_DATA_1'length-1 downto 0);
		--		MEM_MF_DATA_2 <= FIFO_OUT_2(MEM_MF_DATA_1'length-1 downto 0);
		--	end if;
		--end process; -- end FIFO_DATA_MULT

		FIFO_DATA_VALID : process(S_AXIS_ACLK)
		begin
			if rising_edge(S_AXIS_ACLK) then
				if(S_AXIS_ARESETN = '0') then
					MEM_STREAM_DATA_VALID <= (others => '0');
					MEM_OUT_DATA_VALID <= (others => '0');
				else
					MEM_STREAM_DATA_VALID <= (others => '0');
					MEM_OUT_DATA_VALID <= (others => '0');
					for i in 0 to bytesInStream-1 loop
						if to_integer(MEM_ADDR) + i < write_pointer then
							MEM_STREAM_DATA_VALID(i) <= '1';
						end if;

						if to_integer(MEM_OUT_ADDR) + i < write_pointer then
							MEM_OUT_DATA_VALID(i) <= '1';
						end if;
					end loop;
				end if;
			end if;
		end process; -- end FIFO_DATA_VALID
	end block FIFO_READ;


	---- Streaming output data from FIFO as memory
	--MEM_OUT : process(S_AXIS_ACLK)
	--  constant bytesInStream : natural := (C_HASHNATOR_WIDTH+(C_HASHNATOR_COUNT-1)*C_HASHNATOR_OFFSET)/8;
	--	variable index : natural;
	--begin
	--	if rising_edge(S_AXIS_ACLK) then
	--		if (MEM_MF_ADDR_1+bytesInStream-1) < NUMBER_OF_INPUT_WORDS  then
	--			if MEM_MODE = '0' then
	--				MEM_STREAM_DATA_VALID <= (others => '0');
	--				for i in 0 to bytesInStream-1 loop
	--					MEM_STREAM_DATA((i+1)*8-1 downto i*8) <= stream_data_fifo(to_integer(unsigned(MEM_ADDR))+i);
	--					if MEM_ADDR + i < write_pointer then
	--						MEM_STREAM_DATA_VALID(i) <= '1';
	--					end if;
	--				end loop;

	--				for i in 0 to bytesInStream-1 loop
	--					MEM_HASHING_DATA((i+1)*8-1 downto i*8) <= stream_data_fifo(to_integer(unsigned(MEM_ADDR))+i);
	--				end loop;
	--			else
	--				for i in 0 to bytesInStream-1 loop
	--					MEM_MF_DATA_1((i+1)*8-1 downto i*8) <= stream_data_fifo(to_integer(unsigned(MEM_MF_ADDR_1))+i);
	--				end loop;

	--				for i in 0 to bytesInStream-1 loop
	--					MEM_MF_DATA_2((i+1)*8-1 downto i*8) <= stream_data_fifo(to_integer(unsigned(MEM_MF_ADDR_2))+i);
	--				end loop;
	--			end if;
	--		end if;
	--	end if;
	--end process; -- end MEM_OUT

	DATA_READY <= writes_done;
	-- Add user logic here
	DATA_SIGNALS : process(all)
		variable write_pointer_byte : natural range 0 to 2**bit_num-1;
	begin
		write_pointer_byte := write_pointer;
		--DATA_READY <= '0';
		--if (write_pointer > 20) and ((write_pointer - MEM_ADDR) > 20) then
		--	DATA_READY <= '1';
		--end if;
		DATA_EMPTY <= '0';
		DATA_ENDING <= '0';
		if (MEM_OUT_ADDR >= write_pointer_byte) then
			DATA_EMPTY <= '1' and writes_done;
		end if;
		if (write_pointer >= 25 and MEM_ADDR >= write_pointer_byte-20) then
			DATA_ENDING <= '1';
		end if;
	end process; -- end DATA_SIGNALS
	-- User logic ends

end arch_imp;
