library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.all;

entity ByteBram is
  generic(
    C_RAM_WIDTH : integer := 8;
    C_RAM_DEPTH : integer := 512;
    C_RAM_PERFORMANCE : string := "LOW_LATENCY"
  );
  port(
    addra : in std_logic_vector(clogb2(C_RAM_DEPTH)-1 downto 0);                          -- Port A Address bus, width determined from RAM_DEPTH
    addrb : in std_logic_vector(clogb2(C_RAM_DEPTH)-1 downto 0);                          -- Port B Address bus, width determined from RAM_DEPTH
    dina  : in std_logic_vector(C_RAM_WIDTH-1 downto 0);                                  -- Port A RAM input data
    dinb  : in std_logic_vector(C_RAM_WIDTH-1 downto 0);                                  -- Port B RAM input data
    clka  : in std_logic;                                                                 -- Port A Clock
    clkb  : in std_logic;                                                                 -- Port B Clock
    wea   : in std_logic;                                                                 -- Port A Write enable
    web   : in std_logic;                                                                 -- Port B Write enable
    ena   : in std_logic;                                                                 -- Port A RAM Enable, for additional power savings, disable port when not in use
    enb   : in std_logic;                                                                 -- Port B RAM Enable, for additional power savings, disable port when not in use
    rsta  : in std_logic;                                                                 -- Port A Output reset (does not affect memory contents)
    rstb  : in std_logic;                                                                  -- Port B Output register enable
    douta : out  std_logic_vector(C_RAM_WIDTH-1 downto 0);                                  -- Port A RAM output data
    doutb : out  std_logic_vector(C_RAM_WIDTH-1 downto 0)
  );
end entity; -- end ByteBram

architecture Behavioral of ByteBram is
    --Insert the following in the architecture before the begin keyword
    --  The following function calculates the address width based on specified RAM depth
    function clogb2( depth : natural) return integer is
    variable temp    : integer := depth;
    variable ret_val : integer := 0;
    begin
        while temp > 1 loop
            ret_val := ret_val + 1;
            temp    := temp / 2;
        end loop;
      	return ret_val;
    end function;
                                                                    -- Port B Output reset (does not affect memory contents)
    signal regcea: std_logic;                                                                 -- Port A Output register enable
    signal regceb: std_logic;                                                                    -- Port B RAM output data
    signal douta_reg : std_logic_vector(C_RAM_WIDTH-1 downto 0) := (others => '0');           -- Port A RAM output data when RAM_PERFORMANCE = HIGH_PERFORMANCE
    signal doutb_reg : std_logic_vector(C_RAM_WIDTH-1 downto 0) := (others => '0');           -- Port B RAM output data when RAM_PERFORMANCE = HIGH_PERFORMANCE

    type ram_type is array (C_RAM_DEPTH-1 downto 0) of std_logic_vector (C_RAM_WIDTH-1 downto 0);      -- 2D Array Declaration for RAM signal
    signal ram_data_a : std_logic_vector(C_RAM_WIDTH-1 downto 0) ;
    signal ram_data_b : std_logic_vector(C_RAM_WIDTH-1 downto 0) ;

    -- Define RAM
    shared variable ram_name : ram_type := (others => (others => '0'));
begin

  --Insert the following in the architecture after the begin keyword
  process(clka)
  begin
      if(clka'event and clka = '1') then
          if(ena = '1') then
              if(wea = '1') then
                  ram_name(to_integer(unsigned(addra))) := dina;
              else
                  ram_data_a <= ram_name(to_integer(unsigned(addra)));
          end if;
      end if;
  end if;

  end process;

  process(clkb)
  begin
      if(clkb'event and clkb = '1') then
          if(enb = '1') then
              if(web = '1') then
                  ram_name(to_integer(unsigned(addrb))) := dinb;
              else
                  ram_data_b <= ram_name(to_integer(unsigned(addrb)));
              end if;
          end if;
      end if;
  end process;

  --  Following code generates LOW_LATENCY (no output register)
  --  Following is a 1 clock cycle read latency at the cost of a longer clock-to-out timing

  no_output_register : if C_RAM_PERFORMANCE = "LOW_LATENCY" generate
      douta <= ram_data_a;
      doutb <= ram_data_b;
  end generate;

  --  Following code generates HIGH_PERFORMANCE (use output register)
  --  Following is a 2 clock cycle read latency with improved clock-to-out timing

  output_register : if C_RAM_PERFORMANCE = "HIGH_PERFORMANCE"  generate
  process(clka)
  begin
      if(clka'event and clka = '1') then
          if(rsta = '1') then
              douta_reg <= (others => '0');
          elsif(regcea = '1') then
              douta_reg <= ram_data_a;
          end if;
      end if;
  end process;
  douta <= douta_reg;

  process(clkb)
  begin
      if(clkb'event and clkb = '1') then
          if(rstb = '1') then
              doutb_reg <= (others => '0');
          elsif(regceb = '1') then
              doutb_reg <= ram_data_b;
          end if;
      end if;
  end process;
  doutb <= doutb_reg;

  end generate;


end architecture; -- end Behavioral
