-- args: --ieee=synopsys
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.all;
use work.LZ4CompressorLib.all;
use ieee.std_logic_misc.all;


entity LZ4CompressorController is
  generic(
    MEMORY_BUS_STEP : natural := 8
  );
  port(
    clk : in  std_logic;
    rst : in  std_logic;
    ce : in std_logic;
    data_ready : in std_logic;
    data_empty : in std_logic;
    data_ending : in std_logic;
    matcher_ce : out std_logic;
    delay_matcher_ce : out  std_logic;
    finder_list_match : out std_logic;
    finder_match_addr : out LZ4Address;
    current_stream_head : out LZ4Address;
    literal_valid : out std_logic;
    mem_mode : out std_logic;
    mem_addr : out LZ4Address;
    unit_working : out std_logic;
    matcher_match : in  std_logic_vector(MEMORY_BUS_STEP-1 downto 0);
    matcher_match_addr : in std_logic_vector(LZ4Address'length*MEMORY_BUS_STEP-1 downto 0);
    matcher_match_source_addr : in std_logic_vector(LZ4Address'length*MEMORY_BUS_STEP-1 downto 0);
    matcher_rst_pipe : out std_logic;
    finder_done : in std_logic;
    finder_match_length : in LZ4MatchLength;
    finder_current_end_stop : out LZ4Address;
    output_match_length : out LZ4Address;
    output_literal_length : out LZ4Address;
    output_match_offset : out LZ4Address;
    output_valid : out std_logic;
    pipe_override : out std_logic;
    pipe_override_valid : out std_logic_vector(MEMORY_BUS_STEP-1 downto 0)
  );
end entity; -- end LZ4CompressorController

architecture Behavioral of LZ4CompressorController is

  type ControlStateType is (IDLE, INIT, MEM_SERVE, PIPE_DELAY, FILTER_MATCH, PROC_MATCH, WRITE_LIT, WAIT_MATCH, INIT_FINDER, WAIT_FINDER, RST_PIPE_LINE, WRITE_MATCH);
  signal current_state : ControlStateType;
  signal next_state : ControlStateType;

  signal out_matcher_ce          : std_logic;
  signal out_delay_matcher_ce    : std_logic;
  signal out_finder_list_match   : std_logic;
  signal out_finder_match_addr   : LZ4Address;
  signal out_current_stream_head : LZ4Address;
  signal out_mem_mode            : std_logic;
  signal out_mem_addr            : LZ4Address;
  signal out_unit_working        : std_logic;
  signal out_matcher_rst_pipe    : std_logic;
  signal out_literal_valid       : std_logic;


  signal start_mem_addr          : LZ4Address;
  signal current_match_length    : LZ4MatchLength;
  signal current_literal_length  : LZ4MatchLength;
  signal current_match_offset    : LZ4MatchLength;

  signal current_is_match : std_logic;
  signal current_match : std_logic_vector(MEMORY_BUS_STEP-1 downto 0);
  signal current_match_addr : std_logic_vector(LZ4Address'length*MEMORY_BUS_STEP-1 downto 0);
  signal current_match_source_addr : std_logic_vector(LZ4Address'length*MEMORY_BUS_STEP-1 downto 0);
  signal current_proc_addr : LZ4Address;
  signal current_proc_word_end_addr : LZ4Address;
  signal current_proc_addr_index : unsigned(2 downto 0);
  signal current_proc_leading_zeros : unsigned(4 downto 0);

  constant PIP_DELAY_MAX : natural := 8;
  constant PIP_DELAY_ADDR_OFFSET : natural := PIP_DELAY_MAX*MEMORY_BUS_STEP;
  signal pip_delay_done : boolean;
  signal pip_delay_cntr : natural;

  signal debug_tmp_leading_zeros : std_logic_vector(MEMORY_BUS_STEP-1 downto 0);
begin
  current_match_offset <= out_finder_match_addr - out_current_stream_head;
  current_literal_length <= out_finder_match_addr - start_mem_addr;
  current_is_match <= OR_REDUCE(matcher_match);
  current_proc_addr_index <= (current_proc_addr(2 downto 0)-current_proc_word_end_addr(2 downto 0));
  finder_current_end_stop <= (others => '1');
  --finder_current_end_stop <= current_proc_word_end_addr;

  FINDER_ADDRESS_SELECTOR : process(all)
  begin
    out_finder_match_addr <= unsigned(current_match_source_addr((to_integer(current_proc_addr_index)+1)*LZ4Address'length-1 downto to_integer(current_proc_addr_index)*LZ4Address'length));
    out_current_stream_head   <= unsigned(current_match_addr       ((to_integer(current_proc_addr_index)+1)*LZ4Address'length-1 downto to_integer(current_proc_addr_index)*LZ4Address'length));
  end process; -- end FINDER_ADDRESS_SELECTOR

  CONTROL_SEQ : process(clk)
    variable tmp_leading_zeros : std_logic_vector(MEMORY_BUS_STEP-1 downto 0);
  begin
    if rising_edge(clk) then
      if rst = '1' then
        current_state <= IDLE;
        out_mem_addr <= to_unsigned(0, out_mem_addr'length);
        pip_delay_done <= false;
        pip_delay_cntr <= 0;
        start_mem_addr <= (others => '0');
        current_match_length <= (others => '0');
        current_proc_addr <= (others =>'0');
        current_proc_word_end_addr <= (others =>'0');
        out_finder_list_match <= '0';
        current_match_addr <= (others => '0');
        current_match <= (others => '0');
        current_match_source_addr <= (others => '0');
        current_proc_leading_zeros <= (others => '0');
      elsif(ce = '1') then
        current_state <= next_state;
        out_finder_list_match <= '0';

        if(current_is_match = '1') then
          if data_empty = '0' then
            current_proc_addr <= out_mem_addr - PIP_DELAY_ADDR_OFFSET;
            current_proc_word_end_addr <= out_mem_addr + MEMORY_BUS_STEP - PIP_DELAY_ADDR_OFFSET;
            out_mem_addr <= out_mem_addr - PIP_DELAY_ADDR_OFFSET;
          end if;

          current_match_source_addr <= matcher_match_source_addr;
          current_match_addr <= matcher_match_addr;
          current_match <= matcher_match;
        end if;

        if(finder_done = '1') then
          current_match_length <= finder_match_length;
        end if;

        case( next_state ) is
          when others =>
        end case; -- end case next_state

        case( current_state ) is
          when INIT =>
            out_mem_addr <= to_unsigned(0, out_mem_addr'length);
            pip_delay_done <= false;
            pip_delay_cntr <= 0;
            start_mem_addr <= (others => '0');
          when MEM_SERVE =>
            start_mem_addr <= out_mem_addr-current_proc_leading_zeros;
            pip_delay_cntr <= pip_delay_cntr+1;
            if data_empty = '0' then
              out_mem_addr <= out_mem_addr + MEMORY_BUS_STEP;
            end if;
          when PIPE_DELAY =>
            pip_delay_cntr <= pip_delay_cntr+1;
            if pip_delay_cntr = PIP_DELAY_MAX-1 then
              pip_delay_done <= true;
            end if;
            if data_empty = '0' and next_state /= PROC_MATCH   then
              out_mem_addr <= out_mem_addr + MEMORY_BUS_STEP;
            end if;
          when WAIT_MATCH  =>
            if data_empty = '0' and next_state /= PROC_MATCH  then
              out_mem_addr <= out_mem_addr + MEMORY_BUS_STEP;
            end if;
          when FILTER_MATCH =>
            for i in 0 to MEMORY_BUS_STEP-1 loop
              if current_match(i) = '1' and
                  (unsigned(current_match_source_addr((i+1)*LZ4Address'length-1 downto i*LZ4Address'length))
                  - unsigned(current_match_addr((i+1)*LZ4Address'length-1 downto i*LZ4Address'length))) < 4 then
                current_match(i) <= '0';
              end if;
            end loop;
          when PROC_MATCH =>
            if current_proc_addr < current_proc_word_end_addr then
              -- If leading zeros
              if(current_match(to_integer(current_proc_addr_index)) = '0') then
                tmp_leading_zeros := (others => '1');
                for i in 0 to tmp_leading_zeros'length-1 loop
                  if to_integer(current_proc_addr_index)+i < current_match'length then
                    tmp_leading_zeros(i) := current_match(to_integer(current_proc_addr_index)+i);
                  end if;
                end loop;
                debug_tmp_leading_zeros <= tmp_leading_zeros;
                --Docela pruser
                for i in 0 to 7 loop
                  if(tmp_leading_zeros(i) = '1') then
                    exit;
                  else
                    current_proc_leading_zeros <= to_unsigned(i+1, current_proc_leading_zeros'length);
                  end if;
                end loop;

                --current_proc_leading_zeros <= to_unsigned(HighestZero(tmp_leading_zeros), current_proc_leading_zeros'length);
              else
                current_proc_leading_zeros <= (others => '0');
                out_finder_list_match <= '1';
              end if;
            end if;
          when WAIT_FINDER =>
            if finder_done = '1' and finder_match_length < 4 then
              current_match(to_integer(current_proc_addr_index)) <= '0';
            end if;
          when WRITE_LIT =>
            current_proc_addr <= current_proc_addr + current_proc_leading_zeros;
            out_mem_addr <= current_proc_addr + current_proc_leading_zeros;
          when WRITE_MATCH =>
            out_mem_addr <= out_finder_match_addr + current_match_length;
            current_proc_addr <= current_proc_addr+current_match_length;
          when RST_PIPE_LINE =>
            pip_delay_cntr <= 0;
            pip_delay_done <= false;
          when others =>
        end case; -- end case current_state
      end if;
    end if;
  end process; -- end CONTROL_SEQ

--current_state, data_ready,
--    pip_delay_done, matcher_match, current_is_match, finder_done
--
  CONTROL_KOMB : process(all)
  begin
    case( current_state ) is
      when IDLE =>
        if data_ready = '1' then
          next_state <= INIT;
        end if;
      when INIT =>
        next_state <= MEM_SERVE;
      when MEM_SERVE =>
        next_state <= PIPE_DELAY;
      when PIPE_DELAY =>
        if pip_delay_done = true then
          next_state <= WAIT_MATCH;
        end if;
        if current_is_match = '1' then
          next_state <= FILTER_MATCH;
        end if;
      when WAIT_MATCH =>
        --if current_is_match = '1' then
        --  next_state <= FILTER_MATCH;
        --end if;
      when FILTER_MATCH =>
        next_state <= PROC_MATCH;
      when PROC_MATCH =>
        if(current_proc_addr < current_proc_word_end_addr) then
          if(current_match(to_integer(current_proc_addr_index)) = '0') then
            next_state <= WRITE_LIT;
          else
            next_state <= WAIT_FINDER;
          end if;
        else
          next_state <= RST_PIPE_LINE;
        end if;
      when INIT_FINDER =>
        next_state <= WAIT_FINDER;
      when WAIT_FINDER =>
        if finder_done = '1' and finder_match_length >= 4 then
          next_state <= WRITE_MATCH;
        elsif finder_done = '1' then
          next_state <= PROC_MATCH;
        end if;
      when WRITE_LIT =>
        next_state <= PROC_MATCH;
      when WRITE_MATCH =>
        next_state <= PROC_MATCH;
      when RST_PIPE_LINE =>
          next_state <= MEM_SERVE;
      when others =>
        next_state <= next_state;
    end case; -- end case current_state
  end process; -- end CONTROL_KOMB


  CONTROL_OUTPUT : process(current_state)
  begin
    out_matcher_ce          <= '0';
    out_delay_matcher_ce    <= '0';
    out_mem_mode            <= '0';
    out_unit_working        <= '0';
    out_matcher_rst_pipe    <= '0';
    out_literal_valid       <= '0';
    output_valid            <= '0';
    pipe_override           <= '0';
    pipe_override_valid     <= (others => '0');


    -- Tmp
    out_delay_matcher_ce <= '1';

    case( current_state ) is
      when INIT =>
        out_unit_working <= '1';
      when PIPE_DELAY =>
        out_matcher_ce <= '1';
        out_delay_matcher_ce <= '1';
      when WAIT_MATCH =>
        out_matcher_ce <= '1';
        out_delay_matcher_ce <= '1';
        out_literal_valid <= '1';
      when INIT_FINDER =>
        out_mem_mode <= '1';
        --out_literal_valid <= '1';
      when WRITE_LIT =>
        out_literal_valid <= '1';
        pipe_override <= '1';
        pipe_override_valid(to_integer(current_proc_leading_zeros)-1 downto 0) <= (others => '1');
      when WAIT_FINDER =>
        out_mem_mode <= '1';
      when RST_PIPE_LINE =>
        out_matcher_rst_pipe <= '1';
      when WRITE_MATCH =>
        output_valid <= '1';
      when others =>
    end case; -- end case current_state
  end process; -- end CONTROL_OUTPUT

  matcher_ce           <= out_matcher_ce and not current_is_match and not data_ending;
  --matcher_ce           <= out_matcher_ce and not data_ending;
  delay_matcher_ce     <= out_delay_matcher_ce;
  finder_list_match    <= out_finder_list_match;
  finder_match_addr    <= out_finder_match_addr;
  current_stream_head  <= out_current_stream_head;
  mem_mode             <= out_mem_mode;
  mem_addr             <= out_mem_addr;
  unit_working         <= out_unit_working;
  matcher_rst_pipe     <= out_matcher_rst_pipe;
  literal_valid        <= out_literal_valid;
  output_match_length  <= current_match_length;
  output_literal_length <= current_literal_length;
  output_match_offset <= current_match_offset;
end architecture; -- end Behavioral
