
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.all;
use work.LZ4CompressorLib.all;
use ieee.math_real.all;
use work.LZ4CompressorOutBufferLib.all;

entity LZ4Compressor_v0_1_OUT_DATA_AXI is
	generic (
		-- Users to add parameters here

		-- User parameters ends
		-- Do not modify the parameters beyond this line

		-- Width of S_AXIS address bus. The slave accepts the read and write addresses of width C_M_AXIS_TDATA_WIDTH.
		C_M_AXIS_TDATA_WIDTH	: integer	:= 32;
		C_FIFO_LENGTH : natural := 15;
		C_DATA_WIDTH : natural := 88;
		C_BLOCK_LENGTH : natural := 5
	);
	port (
		-- Users to add ports here
		LIT_DATA : in std_logic_vector(C_DATA_WIDTH-1 downto 0);
		LIT_DATA_VALID : in std_logic_vector(C_DATA_WIDTH/8-1 downto 0);
		LIT_DATA_VALID_FLAG : in std_logic;
		LIT_DATA_LAST : in std_logic;
		LIT_LENGTH : in LZ4MatchLength;
	  MATCH_LENGTH : in LZ4MatchLength;
	  MATCH_OFFSET : in LZ4Address;
		LENGTH_VALID : in std_logic;
		OUTPUT_READY : out std_logic;
		M_PROC_READY : out std_logic;
		COMPRESSION_END : out std_logic;
		-- User ports ends
		-- Do not modify the ports beyond this line

		-- Global ports
		M_AXIS_ACLK	: in std_logic;
		--
		M_AXIS_ARESETN	: in std_logic;
		-- Master Stream Ports. TVALID indicates that the master is driving a valid transfer, A transfer takes place when both TVALID and TREADY are asserted.
		M_AXIS_TVALID	: out std_logic;
		-- TDATA is the primary payload that is used to provide the data that is passing across the interface from the master.
		M_AXIS_TDATA	: out std_logic_vector(C_M_AXIS_TDATA_WIDTH-1 downto 0);
		-- TSTRB is the byte qualifier that indicates whether the content of the associated byte of TDATA is processed as a data byte or a position byte.
		M_AXIS_TSTRB	: out std_logic_vector((C_M_AXIS_TDATA_WIDTH/8)-1 downto 0);
		-- TLAST indicates the boundary of a packet.
		M_AXIS_TLAST	: out std_logic;
		-- TREADY indicates that the slave can accept a transfer in the current cycle.
		M_AXIS_TREADY	: in std_logic
	);
end LZ4Compressor_v0_1_OUT_DATA_AXI;

architecture implementation of LZ4Compressor_v0_1_OUT_DATA_AXI is
	component LISCCalculator
	generic (
	  C_OUTPUT_WIDTH : natural := 32;
	  C_INPUT_WIDTH  : natural := 16
	);
	port (
	  clk          : in  STD_LOGIC;
	  ce           : in  STD_LOGIC;
	  rst          : in  STD_LOGIC;
	  input_number : in  STD_LOGIC_VECTOR (C_INPUT_WIDTH-1 downto 0);
	  start        : in  STD_LOGIC;
	  output_lisc  : out STD_LOGIC_VECTOR (C_OUTPUT_WIDTH-1 downto 0);
	  output_valid : out STD_LOGIC_VECTOR (C_OUTPUT_WIDTH/8-1 downto 0);
	  done         : out STD_LOGIC
	);
end component LISCCalculator;

	component LZ4CompressorOutBufferInControl
	generic (
	  C_BYTEFIFO_ADDR_LENGTH  : natural;
	  C_BLOCKFIFO_ADDR_LENGTH : natural;
	  C_DATA_WIDTH            : natural
	);
	port (
	  clk                  : in  std_logic;
	  rst                  : in  std_logic;
    m_proc_ready : out std_logic;
    literal_data_valid   : in std_logic_vector(C_DATA_WIDTH/8-1 downto 0);
    literal_data_valid_flag : in std_logic;
	  literal_length_valid : in  std_logic;
    data_last 					 : in std_logic;
    last_block 				 : out std_logic;
    byte_full 					 : in std_logic;
    header_type : out OutputHeaderType;
	  byte_write_ptr       : out unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    byte_write_word_ptr : out unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
	  block_write_ptr      : out unsigned(C_BLOCKFIFO_ADDR_LENGTH-1 downto 0);
	  last_literal_save    : out unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    literal_length : out unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0)
	);
	end component LZ4CompressorOutBufferInControl;

	component LZ4CompressorOutBufferOutControl
	generic (
	  C_BYTEFIFO_ADDR_LENGTH  : natural;
	  C_BLOCKFIFO_ADDR_LENGTH : natural;
	  C_DATA_WIDTH            : natural;
	  C_BLOCK_LENGTH          : natural
	);
	port (
	  clk                : in  std_logic;
	  rst                : in  std_logic;
	  byte_read_ptr      : out unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
	  block_read_ptr     : out unsigned(C_BLOCKFIFO_ADDR_LENGTH-1 downto 0);
	  block_valid        : in  std_logic;
	  block_type         : in  std_logic;
	  literal_start      : in  unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
	  literal_length     : in  unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    current_block_size : in unsigned(C_BYTEFIFO_ADDR_LENGTH-1 downto 0);
    current_header_type : in OutputHeaderType;
	  match_offset       : in  LZ4Address;
	  match_length       : in  LZ4MatchLength;
	  lisc_input_number  : out LZ4MatchLength;
	  lisc_start         : out std_logic;
	  lisc_done          : in  std_logic;
		last_block				 : in  std_logic;
		blocks_done 			 : out std_logic;
	  literals_valid     : out std_logic_vector(C_DATA_WIDTH/8-1 downto 0);
	  write_literal      : out std_logic;
	  write_lisc         : out std_logic;
	  write_token        : out std_logic;
    write_block_length : out std_logic;
    write_block_reserve : out std_logic;
    block_uncompressed : out std_logic;
    write_done 				 : out std_logic;
	  write_match        : out std_logic;
    write_frame_header : out std_logic;
    write_end_mark 	   : out std_logic;
	  output_buffer_full : in  std_logic;
	  token_output       : out std_logic_vector(7 downto 0);
    output_done 			 : out std_logic
	);
	end component LZ4CompressorOutBufferOutControl;

	component WordBram
	generic (
	  UNALIGNED_WRITE : boolean := false;
	  PORT_WORD_WIDTH : natural := 8;
	  BYTE_WIDTH      : natural := 8;
	  WORDS_COUNT     : natural := 2048;
	  DEBUG_VIEW      : boolean := true
	);
	port (
	  clk        : in  std_logic;
	  rst        : in  std_logic;
	  addr_in    : in  unsigned(integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT))))-1 downto 0);
	  addr_out   : in  unsigned(integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT)))) -1 downto 0);
	  data_in_we : in  std_logic_vector(PORT_WORD_WIDTH-1 downto 0);
	  data_in    : in  std_logic_vector(PORT_WORD_WIDTH*BYTE_WIDTH-1 downto 0);
	  data_out   : out std_logic_vector(PORT_WORD_WIDTH*BYTE_WIDTH-1 downto 0)
	);
	end component WordBram;

	constant BYTE_MEM_WORD_COUNT : natural := 512;
	constant BYTE_MEM_BYTE_WIDTH : natural := C_DATA_WIDTH/8;
	constant BYTE_MEM_ADDR_IN_WIDTH : natural := integer(ceil(log2(real(BYTE_MEM_BYTE_WIDTH*BYTE_MEM_WORD_COUNT))));
	constant BYTE_MEM_ADDR_OUT_WIDTH : natural := integer(ceil(log2(real(BYTE_MEM_BYTE_WIDTH*BYTE_MEM_WORD_COUNT))));

	subtype InnerByteBufferAddress is unsigned(C_FIFO_LENGTH-1 downto 0);
	subtype InnerBlockBufferAddress is unsigned(C_BLOCK_LENGTH-1 downto 0);
	type BLOCK_INFO_TYPE is record
		literal_start : InnerByteBufferAddress;
		literal_length : InnerByteBufferAddress;
		match_offset : LZ4Address;
		match_length : LZ4MatchLength;
		block_type : std_logic;
		current_header_type : OutputHeaderType;
	end record BLOCK_INFO_TYPE;
	constant BLOCK_INFO_TYPE_LENGTH : natural := (InnerByteBufferAddress'length*2+LZ4Address'length+LZ4MatchLength'length+1+5);
	subtype BLOCK_INFO_TYPE_V is std_logic_vector(BLOCK_INFO_TYPE_LENGTH-1 downto 0);

	pure function to_blockVector(x : BLOCK_INFO_TYPE) return BLOCK_INFO_TYPE_V is
		variable tmp : BLOCK_INFO_TYPE_V;
	begin
		tmp(InnerByteBufferAddress'length-1 downto 0) := std_logic_vector(x.literal_start);
		tmp(2*InnerByteBufferAddress'length-1 downto InnerByteBufferAddress'length) := std_logic_vector(x.literal_length);
		tmp((2*InnerByteBufferAddress'length+LZ4Address'length)-1 downto 2*InnerByteBufferAddress'length) := std_logic_vector(x.match_length);
		tmp((2*InnerByteBufferAddress'length+LZ4Address'length+LZ4MatchLength'length)-1 downto 2*InnerByteBufferAddress'length+LZ4Address'length) := std_logic_vector(x.match_offset);
		tmp((2*InnerByteBufferAddress'length+LZ4Address'length+LZ4MatchLength'length)) := x.block_type;
		tmp((2*InnerByteBufferAddress'length+LZ4Address'length+LZ4MatchLength'length)+5 downto (2*InnerByteBufferAddress'length+LZ4Address'length+LZ4MatchLength'length)+1) := std_logic_vector(to_unsigned(OutputHeaderType'pos(x.current_header_type),5));
		return tmp;
	end to_blockVector;

	pure function to_blockinfo(x : BLOCK_INFO_TYPE_V) return BLOCK_INFO_TYPE is
		variable tmp : BLOCK_INFO_TYPE;
	begin
		tmp.literal_start := unsigned(x(InnerByteBufferAddress'length-1 downto 0));
		tmp.literal_length := unsigned(x(2*InnerByteBufferAddress'length-1 downto InnerByteBufferAddress'length));
		tmp.match_length := unsigned(x((2*InnerByteBufferAddress'length+LZ4Address'length)-1 downto 2*InnerByteBufferAddress'length));
		tmp.match_offset := unsigned(x((2*InnerByteBufferAddress'length+LZ4Address'length+LZ4MatchLength'length)-1 downto (2*InnerByteBufferAddress'length+LZ4Address'length)));
		tmp.block_type := x(2*InnerByteBufferAddress'length+LZ4Address'length+LZ4MatchLength'length);
		tmp.current_header_type := OutputHeaderType'val(to_integer(unsigned(x(2*InnerByteBufferAddress'length+LZ4Address'length+LZ4MatchLength'length + 5 downto 2*InnerByteBufferAddress'length+LZ4Address'length+LZ4MatchLength'length+1))));
		return tmp;
	end to_blockinfo;
	attribute ram_style: string;

	type BLOCK_FIFO_TYPE is array(0 to (2**C_BLOCK_LENGTH-1)) of BLOCK_INFO_TYPE_V;
	signal stream_block_fifo : BLOCK_FIFO_TYPE;
	attribute ram_style of stream_block_fifo : signal is "block";

	signal block_read_ptr : InnerBlockBufferAddress;
	signal block_write_ptr : InnerBlockBufferAddress;
	signal block_full : std_logic;
	signal block_empty : std_logic;
	signal block_valid : std_logic;


	signal byte_read_ptr : InnerByteBufferAddress;
	signal byte_write_ptr : InnerByteBufferAddress;
	signal byte_write_word_ptr : InnerByteBufferAddress;
	signal byte_full : std_logic;
	signal byte_empty : std_logic;
	signal byte_output : std_logic_vector(LIT_DATA'length-1 downto 0);
	signal last_literal_save : InnerByteBufferAddress;
	signal literal_length : InnerByteBufferAddress;


	type BYTE_FIFO_TYPE is array (0 to (2**C_FIFO_LENGTH-1)) of std_logic_vector(7 downto 0);
	signal output_fifo : BYTE_FIFO_TYPE;
	attribute ram_style of output_fifo : signal is "block";

	constant OUTPUT_MEM_WORD_COUNT : natural := 512;
	constant OUTPUT_MEM_DATA_WIDTH : natural := MAX(C_M_AXIS_TDATA_WIDTH, C_DATA_WIDTH);
	constant OUTPUT_MEM_BYTE_WIDTH : natural := OUTPUT_MEM_DATA_WIDTH/8;
	constant OUTPUT_MEM_ADDR_IN_WIDTH : natural := integer(ceil(log2(real(OUTPUT_MEM_BYTE_WIDTH*OUTPUT_MEM_WORD_COUNT))));
	constant OUTPUT_MEM_ADDR_OUT_WIDTH : natural := integer(ceil(log2(real(OUTPUT_MEM_BYTE_WIDTH*OUTPUT_MEM_WORD_COUNT))));


	signal output_fifo_in : std_logic_vector(OUTPUT_MEM_DATA_WIDTH-1 downto 0);
	signal output_fifo_out : std_logic_vector(OUTPUT_MEM_DATA_WIDTH-1 downto 0);
	signal output_fifo_in_strb : std_logic_vector(OUTPUT_MEM_DATA_WIDTH/8-1 downto 0);
	signal output_fifo_in_addr : InnerByteBufferAddress;

	signal output_read_ptr : InnerByteBufferAddress;
	signal output_read_mel_ptr : InnerByteBufferAddress;
	signal output_write_ptr : InnerByteBufferAddress;
	signal output_valid_ptr : InnerByteBufferAddress;
	signal output_word_valid : std_logic;
	signal output_word_valid_count : unsigned(C_M_AXIS_TDATA_WIDTH-1 downto 0);
	signal output_fifo_full : std_logic;


	type INPUT_FSM_STATE is (I_IDLE, I_LITERAL_SAVE, I_HEADER_SAVE);
	signal input_current_state : INPUT_FSM_STATE;
	signal input_next_state : INPUT_FSM_STATE;


	type OUTPUT_FSM_STATE is (O_IDLE, O_TOKEN, O_LIT_LISC, O_LITERALS, O_MATCH_LISC);
	signal output_current_state : OUTPUT_FSM_STATE;
	signal output_next_state : OUTPUT_FSM_STATE;
	signal output_current_blockindex : InnerBlockBufferAddress;
	signal output_current_block : BLOCK_INFO_TYPE;
	signal output_literals_index : InnerByteBufferAddress;
	signal current_block_size : InnerByteBufferAddress;

	constant LISC_CALC_INPUT_WIDTH : natural := InnerByteBufferAddress'length;
	constant LISC_CALC_OUTPUT_WIDTH : natural := C_M_AXIS_TDATA_WIDTH;


	signal rst : std_logic;
	signal lisc_calc_input_number : STD_LOGIC_VECTOR (LISC_CALC_INPUT_WIDTH-1 downto 0);
	signal lisc_calc_start        : STD_LOGIC;
	signal lisc_calc_output_lisc  : STD_LOGIC_VECTOR (LISC_CALC_OUTPUT_WIDTH-1 downto 0);
	signal lisc_calc_output_valid : STD_LOGIC_VECTOR (LISC_CALC_OUTPUT_WIDTH/8-1 downto 0);
	signal lisc_calc_done         : STD_LOGIC;


	signal out_M_AXIS_TVALID : std_logic;
	signal out_M_AXIS_TDATA : std_logic_vector(C_M_AXIS_TDATA_WIDTH-1 downto 0);
	signal out_M_AXIS_TSTRB : std_logic_vector((C_M_AXIS_TDATA_WIDTH/8)-1 downto 0);
	signal out_M_AXIS_TLAST : std_logic;
	signal out_OUTPUT_READY : std_logic;

	signal ctrl_lisc_input_number  : LZ4MatchLength;
	signal literals_valid     : std_logic_vector(C_DATA_WIDTH/8-1 downto 0);
	signal write_literal      : std_logic;
	signal write_lisc         : std_logic;
	signal write_token        : std_logic;
	signal write_match        : std_logic;
	signal write_done         : std_logic;
	signal write_block_length : std_logic;
	signal write_frame_header : std_logic;
	signal write_end_mark	    : std_logic;
	signal write_block_reserve : std_logic;
	signal block_uncompressed : std_logic;
	signal token_output       : std_logic_vector(7 downto 0);
	signal last_block 			  : std_logic;
	signal blocks_done			  : std_logic;
	signal output_done	 	 		: std_logic;
	signal inputBlock : BLOCK_INFO_TYPE;
	signal header_type : OutputHeaderType;
	signal current_stream_address : LZ4Address;

	constant LZ4_HEADER : std_logic_vector(8*7-1 downto 0) := X"04224d18604082";
begin
	COMPRESSION_END <= blocks_done;
	rst <= not M_AXIS_ARESETN;
	lisc_calc_input_number <= std_logic_vector(resize(ctrl_lisc_input_number, lisc_calc_input_number'length));

	block_valid <= not block_empty;
	LZ4CompressorOutBufferInControl_i : LZ4CompressorOutBufferInControl
	generic map (
	  C_BYTEFIFO_ADDR_LENGTH  => C_FIFO_LENGTH,
	  C_BLOCKFIFO_ADDR_LENGTH => C_BLOCK_LENGTH,
	  C_DATA_WIDTH            => C_DATA_WIDTH
	)
	port map (
	  clk                  => M_AXIS_ACLK,
	  rst                  => rst,
		m_proc_ready 				 => M_PROC_READY,
	  literal_data_valid   => LIT_DATA_VALID,
	  literal_data_valid_flag   => LIT_DATA_VALID_FLAG,
	  literal_length_valid => LENGTH_VALID,
		data_last 					 => LIT_DATA_LAST,
		byte_full 					 => byte_full,
		last_block 				 => last_block,
		header_type 			 => header_type,
	  byte_write_ptr       => byte_write_ptr,
		byte_write_word_ptr  => byte_write_word_ptr,
	  block_write_ptr      => block_write_ptr,
	  last_literal_save    => last_literal_save,
		literal_length 			 => literal_length
	);

	BYTE_FULL_PROC : process(byte_write_ptr, byte_read_ptr)
	begin
		byte_full <= '0';
		for i in 1 to LIT_DATA'length-1 loop
			if byte_write_ptr+i = byte_read_ptr then
				byte_full <= '1';
				exit;
			end if;
		end loop;
	end process; -- end BYTE_FULL_PROC


	byte_empty <= '1' when byte_write_ptr = byte_read_ptr else '0';
	--FIFO_BYTE_INPUT : process(M_AXIS_ACLK)
	--begin
	--	if rising_edge(M_AXIS_ACLK) then
	--		InFor : for i in 0 to LIT_DATA'length/8-1 loop
	--			if LIT_DATA_VALID(i) = '1' then
	--				stream_data_fifo(i+to_integer(byte_write_ptr)) <= LIT_DATA((i+1)*8-1 downto i*8);
	--			end if;
	--		end loop;  -- end for InFor
	--	end if;
	--end process; -- end FIFO_BYTE_INPUT

	--FIFO_BYTE_OUTPUT : process(M_AXIS_ACLK)
	--begin
	--	if rising_edge(M_AXIS_ACLK) then
	--		InFor : for i in 0 to LIT_DATA'length/8-1 loop
	--			byte_output((i+1)*8-1 downto i*8) <= stream_data_fifo(to_integer(byte_read_ptr)+i);
	--		end loop;  -- end for InFor
	--	end if;
	--end process; -- end FIFO_BYTE_INPUT

	WordBram_i : WordBram
	generic map (
    UNALIGNED_WRITE => true,
	  PORT_WORD_WIDTH  => BYTE_MEM_BYTE_WIDTH,
	  WORDS_COUNT => BYTE_MEM_WORD_COUNT
	)
	port map (
	  clk        => M_AXIS_ACLK,
	  rst        => rst,
	  addr_in    => byte_write_ptr(BYTE_MEM_ADDR_IN_WIDTH-1 downto 0),
	  addr_out   => byte_read_ptr(BYTE_MEM_ADDR_OUT_WIDTH-1 downto 0),
	  data_in_we => LIT_DATA_VALID,
	  data_in    => LIT_DATA,
	  data_out   => byte_output
	);



	block_full <= '1' when (block_write_ptr+1) = block_read_ptr else '0';
	block_empty <= '1' when block_write_ptr = block_read_ptr else '0';
	FIFO_BLOCK_INPUT : process(all)
		variable tmpBlock : BLOCK_INFO_TYPE;
		variable tmpVec : BLOCK_INFO_TYPE_V;
	begin
		tmpBlock.literal_start := last_literal_save;
		if LIT_DATA_LAST = '1' or byte_full = '1' or header_type = WriteUnfinishedSequence then
			tmpBlock.literal_length := resize(byte_write_ptr-last_literal_save, tmpBlock.literal_length'length);
		else
			if header_type = FillBlockSize or header_type = ReserveBlockSize then
				tmpBlock.literal_length := (others => '0');
			else
				tmpBlock.literal_length := resize(literal_length, tmpBlock.literal_length'length);
			end if;
		end if;
		if header_type = FillBlockSize or header_type = ReserveBlockSize or header_type = WriteUnfinishedSequence then
			tmpBlock.match_length := (others => '0');
		else
			tmpBlock.match_length := MATCH_LENGTH;
		end if;
		tmpBlock.match_offset := MATCH_OFFSET;
		tmpBlock.current_header_type := header_type;
		inputBlock <= tmpBlock;
		--tmpBlock.block_type := LIT_DATA_LAST or byte_full;
	end process; -- end FIFO_BLOCK_INPUT

	BLOCK_FIFO_WRITE : process(M_AXIS_ACLK)
	begin
		if rising_edge(M_AXIS_ACLK) then
			stream_block_fifo(to_integer(block_write_ptr)) <= to_blockVector(inputBlock);
		end if;
	end process; -- end BLOCK_FIFO_WRITE


	out_OUTPUT_READY <= not (byte_full or block_full);
	-- FIFO_OUTPUT --
  output_current_block <= to_blockinfo(stream_block_fifo(to_integer(block_read_ptr)));


	WordBram_outMem : WordBram
	generic map (
	  UNALIGNED_WRITE => true,
	  PORT_WORD_WIDTH => OUTPUT_MEM_BYTE_WIDTH,
	  WORDS_COUNT     => OUTPUT_MEM_WORD_COUNT
	)
	port map (
	  clk        => M_AXIS_ACLK,
	  rst        => rst,
	  addr_in    => output_fifo_in_addr(OUTPUT_MEM_ADDR_IN_WIDTH-1 downto 0),
	  addr_out   => output_read_mel_ptr(OUTPUT_MEM_ADDR_OUT_WIDTH-1 downto 0),
	  data_in_we => output_fifo_in_strb,
	  data_in    => output_fifo_in,
	  data_out   => output_fifo_out
	);
	out_M_AXIS_TDATA <= output_fifo_out(out_M_AXIS_TDATA'length-1 downto 0);

	CURRENT_STREAM_HEAD_COUNTER : process(M_AXIS_ACLK)
	begin
		if rising_edge(M_AXIS_ACLK) then
			if rst = '1' then
				current_stream_address <= (others => '0');
			else
				if write_match = '1' then
					current_stream_address <= current_stream_address + output_current_block.literal_length +
						output_current_block.match_length;
				end if;
			end if;
		end if;
	end process; -- end CURRENT_STREAM_HEAD_COUNTER


	AXI_OUT_FIFO_IN_MUX : process(all)
			variable block_length : unsigned(31 downto 0);
	begin
		output_fifo_in <= (others => '0');
		output_fifo_in_strb <= (others => '0');
		output_fifo_in_addr <= output_write_ptr;

		if write_token = '1' then
			output_fifo_in(7 downto 0) <= token_output;
			output_fifo_in_strb(0) <= '1';
		elsif write_match = '1' then
			output_fifo_in(15 downto 0) <= std_logic_vector(current_stream_address + output_current_block.literal_length - output_current_block.match_offset);
			output_fifo_in_strb(1 downto 0) <= (others => '1');
		elsif write_lisc = '1' then
			output_fifo_in(lisc_calc_output_lisc'length-1 downto 0) <= lisc_calc_output_lisc;
			output_fifo_in_strb(lisc_calc_output_valid'length-1 downto 0) <= lisc_calc_output_valid;
		elsif write_literal = '1' then
			output_fifo_in(byte_output'length-1 downto 0) <= byte_output;
			output_fifo_in_strb(literals_valid'length-1 downto 0) <= literals_valid;
		elsif write_block_length = '1' then
			block_length := resize(output_write_ptr - output_valid_ptr - 4, block_length'length);
			block_length(31) := block_uncompressed;
			output_fifo_in_addr <= output_valid_ptr;
			output_fifo_in(31 downto 0) <= std_logic_vector(block_length(31 downto 0));
			output_fifo_in_strb(3 downto 0) <= (others => '1');
		elsif write_frame_header = '1' then
			for i in 0 to LZ4_HEADER'length/8-1 loop
				output_fifo_in((i+1)*8-1 downto i*8) <= LZ4_HEADER((LZ4_HEADER'length/8-i)*8-1 downto (LZ4_HEADER'length/8-(i+1))*8);
			end loop;
			output_fifo_in_strb(LZ4_HEADER'length/8-1 downto 0) <= (others => '1');
		elsif write_end_mark = '1' then
			output_fifo_in(3 downto 0) <= (others => '0');
			output_fifo_in_strb(3 downto 0) <= (others => '1');
		end if;
	end process; -- end AXI_OUT_FIFO_IN_MUX

	AXI_OUT_WRITE_POINTER : process(M_AXIS_ACLK)
		variable block_length : unsigned(31 downto 0);
		variable next_valid_ptr : InnerByteBufferAddress;
		variable next_write_ptr : InnerByteBufferAddress;
		variable tmpVar : InnerByteBufferAddress;
	begin
		if rising_edge(M_AXIS_ACLK) then
			if rst = '1' then
				output_write_ptr <= to_unsigned(0, output_write_ptr'length);
				output_valid_ptr <= (others => '0');
			else
				next_valid_ptr := output_valid_ptr;
				next_write_ptr := output_write_ptr;


				if write_block_reserve = '1' then
					next_write_ptr := next_write_ptr + 4;
				end if;

				if write_token = '1' then
					next_write_ptr := next_write_ptr + 1;
					--output_fifo(to_integer(output_write_ptr)) <= token_output;
					--output_fifo_in(7 downto 0) <= token_output;
					--output_fifo_in_strb(0) <= '1';
				elsif write_match = '1' then
					next_write_ptr := next_write_ptr + 2;
					--output_fifo(to_integer(output_write_ptr)) <= std_logic_vector(output_current_block.match_offset(7 downto 0));
					--output_fifo(to_integer(output_write_ptr+1)) <= std_logic_vector(output_current_block.match_offset(15 downto 8));
					--output_fifo_in(15 downto 0) <= std_logic_vector(output_current_block.match_offset);
					--output_fifo_in_strb(1 downto 0) <= (others => '1');
				elsif write_lisc = '1' then
					--for i in 0 to lisc_calc_output_valid'length-1 loop
					--	if lisc_calc_output_valid(i) /= '1' then
					--		exit;
					--	else
					--		tmpVar := to_unsigned(i+1,tmpVar'length);
					--		output_fifo(to_integer(output_write_ptr + i)) <= lisc_calc_output_lisc((i+1)*8-1 downto i*8);
					--	end if;
					--end loop;

					next_write_ptr := next_write_ptr + HighestOne(lisc_calc_output_valid);
					--output_fifo_in(lisc_calc_output_lisc'length-1 downto 0) <= lisc_calc_output_lisc;
					--output_fifo_in_strb(lisc_calc_output_valid'length-1 downto 0) <= lisc_calc_output_valid;
				elsif write_literal = '1' then
					--tmpVar := (others => '0');
					--for i in 0 to literals_valid'length-1 loop
					--	if literals_valid(i) /= '1' then
					--		exit;
					--	else
					--		tmpVar := to_unsigned(i+1,tmpVar'length);
					--		output_fifo(to_integer(output_write_ptr + i)) <= byte_output((i+1)*8-1 downto i*8);
					--	end if;
					--end loop;
					next_write_ptr := next_write_ptr + HighestOne(literals_valid);
					--report "Input: " & integer'image(to_integer(unsigned(literals_valid))) & " Output: " & natural'image(HighestOne(literals_valid));
					--output_fifo_in(byte_output'length-1 downto 0) <= byte_output;
					--output_fifo_in_strb(literals_valid'length-1 downto 0) <= literals_valid;
				elsif write_block_length = '1' then
					block_length := resize(output_write_ptr - output_valid_ptr - 4, block_length'length);
					block_length(31) := block_uncompressed;
					--for i in 0 to 3 loop
					--	output_fifo(to_integer(output_valid_ptr+i)) <= std_logic_vector(block_length((i+1)*8-1 downto i*8));
					--end loop;
					--output_fifo_in(31 downto 0) <= std_logic_vector(block_length(31 downto 0));
					--output_fifo_in_strb(3 downto 0) <= (others => '1');
				elsif write_frame_header = '1' then
					next_write_ptr := next_write_ptr + LZ4_HEADER'length/8;
					next_valid_ptr := next_valid_ptr + LZ4_HEADER'length/8;
					--for i in 0 to LZ4_HEADER'length/8-1 loop
					----	output_fifo(to_integer(output_write_ptr+(LZ4_HEADER'length/8-1 - i))) <= LZ4_HEADER((i+1)*8-1 downto i*8);
					--	output_fifo_in((i+1)*8-1 downto i*8) <= LZ4_HEADER((LZ4_HEADER'length/8-i)*8-1 downto (LZ4_HEADER'length/8-(i+1))*8);
					--end loop;
					--output_fifo_in_strb(LZ4_HEADER'length/8-1 downto 0) <= (others => '1');
				elsif write_end_mark = '1' then
					next_write_ptr := next_write_ptr + 4;
					--for i in 0 to 3 loop
					--	output_fifo(to_integer(output_write_ptr+i)) <= (others => '0');
					--end loop;
					--output_fifo_in(3 downto 0) <= (others => '0');
					--output_fifo_in_strb(3 downto 0) <= (others => '1');
				end if;

				if write_done = '1' then
					next_valid_ptr := next_write_ptr;
				end if;
				output_valid_ptr <= next_valid_ptr;
				output_write_ptr <= next_write_ptr;
			end if;
		end if;
	end process; -- end AXI_OUT_FIFO_WRITE


	--AXI_OUTPUT : process(M_AXIS_ACLK)
	--begin
	--	if rising_edge(M_AXIS_ACLK) then
	--		if rst = '1' then
	--			out_M_AXIS_TDATA <= (others => '0');
	--		else
	--			for i in 0 to C_M_AXIS_TDATA_WIDTH/8-1 loop
	--				out_M_AXIS_TDATA(8*(i+1)-1 downto i*8) <= output_fifo(to_integer(output_read_ptr+i));
	--			end loop;
	--		end if;
	--	end if;
	--end process; -- end AXI_OUTPUT


	AXI_OUT_WORD_FLAG : process(output_read_mel_ptr, output_write_ptr, output_valid_ptr)
	begin
		output_word_valid <= '1';
		for i in 0 to C_M_AXIS_TDATA_WIDTH/8-1 loop
			output_word_valid_count <= to_unsigned(i, output_word_valid_count'length);
			if output_read_mel_ptr+i = output_valid_ptr then
				output_word_valid <= '0';
				exit;
			end if;
		end loop;
	end process; -- end BYTE_FULL_PROC


	AXI_OUT_FULL_FLAG : process(output_write_ptr, output_read_ptr)
	begin
		output_fifo_full <= '0';
		for i in 1 to LIT_DATA'length-1 loop
			if output_write_ptr+i = output_read_ptr then
				output_fifo_full <= '1';
				exit;
			end if;
		end loop;
	end process; -- end BYTE_FULL_PROC

	OUTPUT_READ_PTR_MEALY : process(all)
	begin
		if(M_AXIS_TREADY = '1') then
			output_read_mel_ptr <= output_read_ptr+C_M_AXIS_TDATA_WIDTH/8;
		else
			output_read_mel_ptr <= output_read_ptr;
		end if;
	end process; -- end OUTPUT_READ_PTR_MEALY

	current_block_size <= output_write_ptr - output_valid_ptr - 4;
	AXI_OUT_FIFO_READ : process(M_AXIS_ACLK)
	begin
		if rising_edge(M_AXIS_ACLK) then
			if rst = '1' then
				output_read_ptr <= (others => '0');
				out_M_AXIS_TLAST <= '0';
				out_M_AXIS_TVALID <= '0';
				out_M_AXIS_TSTRB <= (others => '0');
			else
				out_M_AXIS_TLAST <= '0';
				out_M_AXIS_TVALID <= '0';
				out_M_AXIS_TSTRB <= (others => '0');

				if output_word_valid = '1' then
					out_M_AXIS_TVALID <= '1';
					out_M_AXIS_TSTRB <= (others => '1');
					if M_AXIS_TREADY = '1' then
						output_read_ptr <= output_read_ptr+C_M_AXIS_TDATA_WIDTH/8;
					end if;

					if blocks_done = '1' and to_integer(output_read_mel_ptr+(C_M_AXIS_TDATA_WIDTH/8)) >= to_integer(output_valid_ptr) then
						out_M_AXIS_TLAST <= '1';
					end if;
				elsif blocks_done = '1' and output_read_mel_ptr /= output_valid_ptr then
					out_M_AXIS_TVALID <= '1';
					out_M_AXIS_TLAST <= '1';
					out_M_AXIS_TSTRB <= (others => '0');
					for i in 0 to C_M_AXIS_TDATA_WIDTH/8-1 loop
						if output_read_mel_ptr + i < output_valid_ptr then
							out_M_AXIS_TSTRB(i) <= '1';
							if M_AXIS_TREADY = '1' then
								output_read_ptr <= output_read_ptr + i + 1;
							end if;
						end if;
					end loop;
				end if;
			end if;
		end if;
	end process; -- end AXI_OUT_FIFO_READ

	LZ4CompressorOutBufferOutControl_i : LZ4CompressorOutBufferOutControl
	generic map (
	  C_BYTEFIFO_ADDR_LENGTH  => C_FIFO_LENGTH,
	  C_BLOCKFIFO_ADDR_LENGTH => C_BLOCK_LENGTH,
	  C_DATA_WIDTH            => C_DATA_WIDTH,
	  C_BLOCK_LENGTH          => BLOCK_INFO_TYPE_V'length
	)
	port map (
	  clk                => M_AXIS_ACLK,
	  rst                => rst,
	  byte_read_ptr      => byte_read_ptr,
	  block_read_ptr     => block_read_ptr,
	  block_valid        => block_valid,
	  block_type         => output_current_block.block_type,
	  literal_start      => output_current_block.literal_start,
	  literal_length     => output_current_block.literal_length,
	  match_offset       => output_current_block.match_offset,
	  match_length       => output_current_block.match_length,
		current_header_type => output_current_block.current_header_type,
		current_block_size => current_block_size,
	  lisc_input_number  => ctrl_lisc_input_number,
	  lisc_start         => lisc_calc_start,
	  lisc_done          => lisc_calc_done,
	  literals_valid     => literals_valid,
		last_block 				 => last_block,
		blocks_done 			 => blocks_done,
	  write_literal      => write_literal,
	  write_lisc         => write_lisc,
	  write_token        => write_token,
	  write_match        => write_match,
		write_frame_header => write_frame_header,
		write_end_mark	   => write_end_mark,
		write_block_reserve => write_block_reserve,
	  output_buffer_full => output_fifo_full,
	  token_output       => token_output,
		write_block_length => write_block_length,
		block_uncompressed => block_uncompressed,
    write_done 				 => write_done,
		output_done 			 => output_done
	);


	LISCCalculator_i : LISCCalculator
	generic map (
	  C_OUTPUT_WIDTH => LISC_CALC_OUTPUT_WIDTH,
	  C_INPUT_WIDTH  => LISC_CALC_INPUT_WIDTH
	)
	port map (
	  clk          => M_AXIS_ACLK,
		ce 				   => '1',
	  rst          => rst,
	  input_number => lisc_calc_input_number,
	  start        => lisc_calc_start,
	  output_lisc  => lisc_calc_output_lisc,
		output_valid => lisc_calc_output_valid,
	  done         => lisc_calc_done
	);


	-- User logic ends
	M_AXIS_TVALID <= out_M_AXIS_TVALID;
	M_AXIS_TDATA <= out_M_AXIS_TDATA;
	M_AXIS_TSTRB <= out_M_AXIS_TSTRB;
	M_AXIS_TLAST <= out_M_AXIS_TLAST;
	OUTPUT_READY <= out_OUTPUT_READY;
end implementation;
