library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;
use work.LZ4CompressorLib.all;
use work.all;

entity MatchLengthFinder is
  generic(
    MEMORY_BUS_WIDTH : natural;
    MATCH_COUNT : natural := 8
  );
  port(
    clk : in  std_logic;
    rst : in  std_logic;
    ce : in std_logic;
    list_match : in  std_logic;
    match_address : in    LZ4Address;
    current_address : in  LZ4Address;
    consecutive_addresses : in std_logic_vector(MATCH_COUNT-1 downto 0);
    mf_address_1 : out LZ4Address;
    mf_address_2 : out LZ4Address;
    mf_end_stop : in LZ4Address;
    mf_start_stop : in LZ4Address;
    mf_data_1 : in std_logic_vector(MEMORY_BUS_WIDTH-1 downto 0);
    mf_data_2 : in std_logic_vector(MEMORY_BUS_WIDTH-1 downto 0);
    match_length : out std_logic_vector(MATCH_COUNT*LZ4MatchLength'length-1 downto 0);
    match_length_valid : out std_logic_vector(MATCH_COUNT-1 downto 0);
    out_match_address : out LZ4Address;
    out_current_address : out LZ4Address;
    length_done : out std_logic
  );
end entity; -- end MatchLengthFinder

architecture Behavioral of MatchLengthFinder is
  constant MEMORY_BUS_BYTES : natural := MEMORY_BUS_WIDTH/8;

  type ControlStateType is (IDLE, ONE_MATCH, MORE_IDLE);
  signal current_state : ControlStateType;
  signal next_state : ControlStateType;
  signal prev_state : ControlStateType;

  signal current_match_addr : LZ4Address;
  signal stream_head_address : LZ4Address;
  signal match_head_address : LZ4Address;
  signal match_error_bus : std_logic_vector(MATCH_COUNT-1 downto 0);
  signal match_error_bus_reg : std_logic_vector(MATCH_COUNT-1 downto 0);
  signal match_error : std_logic;
  signal curr_consecutive_addresses : std_logic_vector(MATCH_COUNT-1 downto 0);
  signal curr_match_address : LZ4Address;
  signal curr_current_address : LZ4Address;

  type MatchLenBus is array (0 to MATCH_COUNT-1) of LZ4MatchLength;
  signal current_match_length : MatchLenBus;
  signal current_match_length_komb : MatchLenBus;
  signal in_mem_passthourg : boolean;
  signal longer_match : boolean;
  signal start_error : std_logic;
begin


CTRL_SEQ : process(clk)
begin
  if rising_edge(clk) then
    if rst = '1' then
      current_state <= IDLE;
      stream_head_address <= (others => '0');
      match_head_address <= (others => '0');
      curr_consecutive_addresses <= (others => '0');
      curr_match_address <= (others => '0');
      curr_current_address <= (others => '0');
    elsif(ce = '1') then
      current_state <= next_state;
      prev_state <= current_state;
      match_error_bus_reg <= match_error_bus;

      case( current_state ) is
        when IDLE =>
          current_match_length <= (others => (others => '0'));
          stream_head_address <= current_address+MEMORY_BUS_BYTES;
          match_head_address <= match_address+MEMORY_BUS_BYTES;
          curr_consecutive_addresses <= consecutive_addresses;
          curr_match_address <= match_address;
          curr_current_address <= current_address;
          longer_match <= false;
        when ONE_MATCH =>
          if match_error = '1' or start_error = '1' then
            curr_consecutive_addresses <= consecutive_addresses;
            curr_match_address <= match_address;
            curr_current_address <= current_address;
            stream_head_address <= current_address+MEMORY_BUS_BYTES;
            match_head_address <= match_address+MEMORY_BUS_BYTES;
            current_match_length <= (others => (others => '0'));
            longer_match <= false;
          else
            if longer_match = true then
              stream_head_address <= stream_head_address+MEMORY_BUS_BYTES;
              match_head_address <= match_head_address+MEMORY_BUS_BYTES;
            end if;

            for i in 0 to MATCH_COUNT-1 loop
              current_match_length(i) <= current_match_length_komb(i);
            end loop;
            longer_match <= true;
          end if;
        when others =>
      end case; -- end case current_state
    end if;
  end if;
end process; -- end CTRL_SEQ

MATCH_LENGTH_KOMB : process(all)
begin
  current_match_length_komb <= (others => (others => '0'));
  for z in 0 to MATCH_COUNT-1 loop
    for i in 0 to MEMORY_BUS_BYTES-1 loop
      if longer_match = false then
        if  i >= z then
          if mf_data_1((i+1)*8-1 downto i*8) = mf_data_2((i+1)*8-1 downto i*8) and
              stream_head_address+i < mf_end_stop then
            current_match_length_komb(z) <= current_match_length(z) + to_unsigned(i-z, current_match_length'length)+1;
          else
            exit;
          end if;
        end if;
      else
        if mf_data_1((i+1)*8-1 downto i*8) = mf_data_2((i+1)*8-1 downto i*8) and
            stream_head_address+i < mf_end_stop then
          current_match_length_komb(z) <= current_match_length(z) + to_unsigned(i, current_match_length'length)+1;
        else
          exit;
        end if;
      end if;
    end loop;
  end loop;
end process; -- end MATCH_LENGTH_KOMB

MATCH_ERROR_PROC : process(all)
begin
  if longer_match = true then
    match_error_bus <= match_error_bus_reg;
  else
    match_error_bus <= (others => '0');
  end if;
  for z in 0 to MATCH_COUNT-1 loop
    for i in 0 to MEMORY_BUS_BYTES-1 loop
      if  i >= z or longer_match = true then
        if mf_data_1((i+1)*8-1 downto i*8) = mf_data_2((i+1)*8-1 downto i*8) and
            stream_head_address+i < mf_end_stop then
        else
          match_error_bus(z) <= '1';
          exit;
        end if;
      end if;
    end loop;
  end loop;
end process; -- end MATCH_ERROR_PROC
match_error <= OR_REDUCE(match_error_bus);

CTRL_KOMB : process(all)
begin
  case( current_state ) is
    when IDLE =>
      if list_match = '1' then
        next_state <= ONE_MATCH;
      else
        next_state <= IDLE;
      end if;
    when ONE_MATCH =>
      if match_error = '1' or start_error = '1' then
        if list_match = '1' then
          next_state <= ONE_MATCH;
        else
          next_state <= IDLE;
        end if;
      else
        next_state <= MORE_IDLE;
      end if;
    when MORE_IDLE =>
      next_state <= ONE_MATCH;
    when others =>
      next_state <= IDLE;
  end case; -- end case current_state
end process; -- end CTRL_KOMB

SKIP_ERROR_PROCESS : process(all)
begin
  start_error <= '0';
  --if curr_current_address < mf_start_stop or (in_mem_passthourg = true and current_address < mf_start_stop) then
  if curr_current_address < mf_start_stop or (((prev_state = IDLE or prev_state = ONE_MATCH) and current_state /= MORE_IDLE) and current_address < mf_start_stop) then
    start_error <= '1';
  end if;
end process; -- end SKIP_ERROR_PROCESS

CTRL_OUTPUT : process(all)
begin
  length_done <= '0';
  match_length <= (others => '0');
  match_length_valid <= (others => '0');
  in_mem_passthourg <= false;
  out_match_address <= (others => '0');
  out_current_address <= (others => '0');
  case( current_state ) is
    when IDLE | ONE_MATCH =>
      in_mem_passthourg <= true;
    when others =>
  end case; -- end case current_state

  if current_state = ONE_MATCH then
    if match_error = '0' and start_error = '0' then
      in_mem_passthourg <= false;
    else
      length_done <= '1';
      if start_error = '0' then
        out_match_address <= curr_match_address;
        out_current_address <= curr_current_address;
        for i in 0 to MATCH_COUNT-1 loop
          if curr_consecutive_addresses(i) = '1' then
            if(to_integer(current_match_length_komb(i)) >= 4) then
              match_length_valid(i) <= '1';
              match_length((i+1)*LZ4MatchLength'length-1 downto i*LZ4MatchLength'length) <= std_logic_vector(current_match_length_komb(i));
            end if;
          end if;
        end loop;
      end if;
    end if;
  end if;
end process; -- end CTRL_OUTPUT

MF_ADDRESS_PROC : process(all)
begin
  mf_address_1 <= (others => '0');
  mf_address_2 <= (others => '0');
  if in_mem_passthourg = true then
    mf_address_1 <= current_address;
    mf_address_2 <= match_address;
  else
    mf_address_1 <= stream_head_address;
    mf_address_2 <= match_head_address;
  end if;
end process; -- end MF_ADDRESS_PROC
end architecture; -- end Behavioral
