library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;
use ieee.math_real.all;
use work.all;
use work.LZ4CompressorLib.all;

entity MatchProcessor is
  generic(
    C_MATCH_COUNT : natural := 8;
    C_MEMORY_BUS_WIDTH : natural := 128
  );
  port(
    clk : in  std_logic;
    rst : in  std_logic;
    out_ce : in std_logic;
    matcher_done : in std_logic;
    memory_length : in LZ4Address;

    match_valid : in std_logic_vector(C_MATCH_COUNT-1 downto 0);
    match_address : in std_logic_vector(C_MATCH_COUNT*LZ4Address'length-1 downto 0);
    match_source_address : in std_logic_vector(C_MATCH_COUNT*LZ4Address'length-1 downto 0);
    finder_mf_data_1 : in std_logic_vector(C_MEMORY_BUS_WIDTH-1 downto 0);
    finder_mf_data_2 : in std_logic_vector(C_MEMORY_BUS_WIDTH-1 downto 0);
    finder_mf_address_1 : out LZ4Address;
    finder_mf_address_2 : out LZ4Address;
    last_serve : out std_logic;

    out_literal_address : out LZ4Address;
    out_litetal_data_valid : out std_logic_vector(C_MEMORY_BUS_WIDTH/8-1 downto 0);
    out_literal_length : out LZ4MatchLength;
    out_match_length : out LZ4MatchLength;
    out_match_offset : out LZ4Address;
    out_match_valid : out std_logic;
    out_match_ready : in std_logic
  );
end entity; -- end MatchProcessor

architecture Behavioral of MatchProcessor is
  component MatchLengthFinder
  generic (
    MEMORY_BUS_WIDTH : natural;
    MATCH_COUNT : natural := 8
  );
  port (
    clk             : in  std_logic;
    rst             : in  std_logic;
    ce              : in  std_logic;
    list_match      : in  std_logic;
    match_address   : in  LZ4Address;
    current_address : in  LZ4Address;
    consecutive_addresses : in std_logic_vector(MATCH_COUNT-1 downto 0);
    mf_address_1    : out LZ4Address;
    mf_address_2    : out LZ4Address;
    mf_end_stop     : in  LZ4Address;
    mf_start_stop   : in LZ4Address;
    mf_data_1       : in  std_logic_vector(MEMORY_BUS_WIDTH-1 downto 0);
    mf_data_2       : in  std_logic_vector(MEMORY_BUS_WIDTH-1 downto 0);
    match_length    : out std_logic_vector(MATCH_COUNT*LZ4MatchLength'length-1 downto 0);
    match_length_valid : out std_logic_vector(MATCH_COUNT-1 downto 0);
    out_match_address : out LZ4Address;
    out_current_address : out LZ4Address;
    length_done     : out std_logic
  );
  end component MatchLengthFinder;

  component ShiftRegister
  generic (
    regDepth  : natural := 6;
    dataWidth : natural := 32
  );
  port (
    clk  : in  STD_LOGIC;
    rst  : in  STD_LOGIC;
    ce   : in  STD_LOGIC;
    din  : in  STD_LOGIC_VECTOR (dataWidth-1 downto 0);
    dout : out STD_LOGIC_VECTOR (dataWidth-1 downto 0)
  );
  end component ShiftRegister;

  component MatchProcessorController
  generic (
    C_MATCH_COUNT      : natural := 8;
    C_MEMORY_BUS_WIDTH : natural := 128
  );
  port (
    clk                 : in  std_logic;
    rst                 : in  std_logic;
    in_ce               : in std_logic;
    out_ce              : in std_logic;
    match_done          : in std_logic;
    memory_length       : in LZ4Address;
    match_length        : in  std_logic_vector(C_MATCH_COUNT*LZ4MatchLength'length-1 downto 0);
    match_length_valid  : in  std_logic_vector(C_MATCH_COUNT-1 downto 0);
    out_match_address   : in  LZ4Address;
    out_current_address : in  LZ4Address;
    last_serve : out std_logic;
    processed_match_address : out LZ4Address;

    out_literal_address : out LZ4Address;
    out_litetal_data_valid : out std_logic_vector(C_MEMORY_BUS_WIDTH/8-1 downto 0);
    out_literal_length : out LZ4MatchLength;
    out_match_length : out LZ4MatchLength;
    out_match_offset : out LZ4Address;
    out_match_valid : out std_logic
  );
  end component MatchProcessorController;


  component WordBram
  generic (
    UNALIGNED_WRITE : boolean := false;
    PORT_WORD_WIDTH : natural := 8;
    BYTE_WIDTH      : natural := 8;
    WORDS_COUNT     : natural := 2048;
    DEBUG_VIEW      : boolean := true;
    INIT_FILE       : string  := ""
  );
  port (
    clk        : in  std_logic;
    rst        : in  std_logic;
    addr_in    : in  unsigned(integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT))))-1 downto 0);
    addr_out   : in  unsigned(integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT)))) -1 downto 0);
    data_in_we : in  std_logic_vector(PORT_WORD_WIDTH-1 downto 0);
    data_in    : in  std_logic_vector(PORT_WORD_WIDTH*BYTE_WIDTH-1 downto 0);
    data_out   : out std_logic_vector(PORT_WORD_WIDTH*BYTE_WIDTH-1 downto 0)
  );
  end component WordBram;


	constant NUMBER_OF_INPUT_WORDS  : integer := 1024/4;
	constant ADDR_IN_WIDTH : natural := integer(ceil(log2(real(C_MATCH_COUNT*NUMBER_OF_INPUT_WORDS))));
	constant ADDR_OUT_WIDTH : natural := integer(ceil(log2(real(C_MATCH_COUNT*NUMBER_OF_INPUT_WORDS))));

  signal matchFifo_read_ptr  : unsigned(ADDR_IN_WIDTH-1 downto 0);
  signal matchFifo_read_ptr_reg  : unsigned(ADDR_IN_WIDTH-1 downto 0);
  signal matchFifo_write_ptr : unsigned(ADDR_OUT_WIDTH-1 downto 0);
  signal matchFifo_empty : std_logic;
  signal matchFifo_empty_reg : std_logic;

  signal current_bus_match_addr : std_logic_vector(C_MATCH_COUNT*LZ4Address'length-1 downto 0);
  signal current_bus_match_source_addr : std_logic_vector(C_MATCH_COUNT*LZ4Address'length-1 downto 0);

  signal current_match_addr : LZ4Address;
  signal current_match_source_addr : LZ4Address;

  signal current_stream_addr : LZ4Address;
  signal current_skip_constant : natural range 0 to C_MATCH_COUNT;
  signal current_skip_constant_reg : natural range 0 to C_MATCH_COUNT;

  signal finder_ce              : std_logic;
  signal finder_list_match      : std_logic;
  signal finder_mf_end_stop     : LZ4Address;
  signal finder_length_done     : std_logic;
  signal finder_consecutive_addresses  : std_logic_vector(C_MATCH_COUNT-1 downto 0);
  signal finder_consecutive_addresses_reg  : std_logic_vector(C_MATCH_COUNT-1 downto 0);

  signal finder_match_length    : std_logic_vector(C_MATCH_COUNT*LZ4MatchLength'length-1 downto 0);
  signal finder_match_length_valid  : std_logic_vector(C_MATCH_COUNT-1 downto 0);
  signal finder_out_match_address   : LZ4Address;
  signal finder_out_current_address : LZ4Address;

  signal finder_match_length_off    : std_logic_vector(C_MATCH_COUNT*LZ4MatchLength'length-1 downto 0);
  signal finder_match_length_valid_off  : std_logic_vector(C_MATCH_COUNT-1 downto 0);
  signal finder_out_match_address_off   : LZ4Address;
  signal finder_out_current_address_off : LZ4Address;
  signal finder_done : std_logic;

  signal control_processed_match_address : LZ4Address;
begin
  finder_mf_end_stop <= memory_length-13;




  --data_in_we => match_valid,
  WordBram_MatchAddrMem : WordBram
  generic map (
    UNALIGNED_WRITE => true,
    PORT_WORD_WIDTH  => C_MATCH_COUNT,
    BYTE_WIDTH => LZ4Address'length,
    WORDS_COUNT => NUMBER_OF_INPUT_WORDS,
    INIT_FILE => "/tmp/matches_lower.hex"
  )
  port map (
    clk        => clk,
    rst        => rst,
    addr_in    => matchFifo_write_ptr,
    addr_out   => matchFifo_read_ptr,
    data_in_we => (others => '0'),
    data_in    => match_address,
    data_out   => current_bus_match_addr
  );
  current_match_addr <= unsigned(current_bus_match_addr((HighestOne(finder_consecutive_addresses_reg))*LZ4Address'length-1 downto (HighestOne(finder_consecutive_addresses_reg)-1)*LZ4Address'length));

  WordBram_MatchSourceAddrMem : WordBram
  generic map (
    UNALIGNED_WRITE => true,
    PORT_WORD_WIDTH  => C_MATCH_COUNT,
    BYTE_WIDTH => LZ4Address'length,
    WORDS_COUNT => NUMBER_OF_INPUT_WORDS,
    INIT_FILE => "/tmp/matches_higher.hex"
  )
  port map (
    clk        => clk,
    rst        => rst,
    addr_in    => matchFifo_write_ptr,
    addr_out   => matchFifo_read_ptr,
    data_in_we => (others => '0'),
    data_in    => match_source_address,
    data_out   => current_bus_match_source_addr
  );
  --Chyba je asi tady
  --current_match_source_addr <= unsigned(current_bus_match_source_addr((HighestOne(finder_consecutive_addresses_reg))*LZ4Address'length-1 downto (HighestOne(finder_consecutive_addresses_reg)-1)*LZ4Address'length));
  current_match_source_addr <= unsigned(current_bus_match_source_addr(LZ4Address'length-1 downto 0));

  EMPTY_PROCESS : process(all)
  begin
    matchFifo_empty <= '0';
    for i in 0 to C_MATCH_COUNT-1 loop
      if matchFifo_write_ptr+i = matchFifo_read_ptr+HighestOne(finder_consecutive_addresses_reg)-1 then
        matchFifo_empty <= '1';
      end if;
    end loop;
  end process; -- end EMPTY_PROCESS


  FIFO_INPUT_PROCESS : process(clk)
  begin
    if rising_edge(clk) then
      if rst = '1' then
        matchFifo_write_ptr <= (others => '0');
        matchFifo_write_ptr <= to_unsigned(1941, ADDR_OUT_WIDTH) after 5 us;
      else
        if(or_reduce(match_valid) = '1') then
          --matchFifo_write_ptr <= matchFifo_write_ptr + HighestOne(match_valid);
        end if;
      end if;
    end if;
  end process; -- end FIFO_INPUT_PROCESS

  SKIP_LOGIC : process(all)
  begin
    current_skip_constant <= 0;
    for i in 0 to C_MATCH_COUNT-1 loop
      if unsigned(current_bus_match_source_addr((I+1)*LZ4Address'length-1 downto I*LZ4Address'length)) < control_processed_match_address then
        current_skip_constant <= i+1;
      end if;
    end loop;
  end process; -- end SKIP_LOGIC

  MATCH_PROCESSOR_CONTROLLER : block
    type ControlStateType is (IDLE, MEM_INIT, INIT_FINDER, WAIT_FINDER, SKIP_MATCHES);
    signal current_state : ControlStateType;
    signal next_state : ControlStateType;

    signal next_match_read_ptr : unsigned(ADDR_IN_WIDTH-1 downto 0);
    signal last_match_read_ptr : unsigned(ADDR_IN_WIDTH-1 downto 0);
    signal idle_skip : boolean;
  begin
      finder_done <= '1' when (matchFifo_empty = '1' and current_state = IDLE and matcher_done = '1') else '0';

      NEXT_ADDRESS : process(all)
        variable step : unsigned(ADDR_IN_WIDTH-1 downto 0);
        variable tmp : unsigned(ADDR_IN_WIDTH-1 downto 0);
      begin
        step := MAX(to_unsigned(HighestOne(finder_consecutive_addresses), ADDR_IN_WIDTH),
            to_unsigned(current_skip_constant, ADDR_IN_WIDTH));
        if step > 2 then
          tmp := matchFifo_read_ptr_reg + step;
        else
          tmp := matchFifo_read_ptr + 1;
        end if;
        --next_match_read_ptr <= tmp;
        if matchFifo_read_ptr <= matchFifo_write_ptr then
            next_match_read_ptr <= MIN(tmp, matchFifo_write_ptr);
        else
            next_match_read_ptr <= MAX(tmp, matchFifo_write_ptr);
        end if;
      end process; -- end NEXT_ADDRESS

      MATCH_PROC_CTRL_SEQ : process(clk)
      begin
        if rising_edge(clk) then
          if rst = '1' then
            matchFifo_read_ptr <= (others => '0');
            current_state <= IDLE;
            matchFifo_empty_reg <= '0';
            --current_skip_constant_reg <= 0;
            finder_consecutive_addresses_reg <= ((1) => '1', others => '0');
            matchFifo_read_ptr_reg <= (others => '0');
          elsif(out_ce = '1') then
            matchFifo_empty_reg <= matchFifo_empty;
            --current_skip_constant_reg <= current_skip_constant;
            finder_consecutive_addresses_reg <= finder_consecutive_addresses;
            matchFifo_read_ptr_reg <= matchFifo_read_ptr;
            current_state <= next_state;
            if(current_state = IDLE and next_state = WAIT_FINDER) then
              matchFifo_read_ptr <= next_match_read_ptr;
            end if;

            case( next_state ) is
              when MEM_INIT =>
                matchFifo_read_ptr <= next_match_read_ptr;
              when others =>
            end case; -- end case next_state

            case( current_state ) is
              when IDLE =>
              when MEM_INIT =>
                matchFifo_read_ptr <= next_match_read_ptr;
              when WAIT_FINDER =>
                if finder_length_done = '1' then
                  matchFifo_read_ptr <= next_match_read_ptr;
                else
                  matchFifo_read_ptr <= matchFifo_read_ptr_reg;
                  finder_consecutive_addresses_reg <= finder_consecutive_addresses_reg;
                end if;
              when others =>
            end case; -- end case current_state
          end if;
        end if;
      end process; -- en  MATCH_PROC_CTRL_SEQ

      MATCH_PROC_CTRL_KOMB_NEXT : process(all)
      begin
        case( current_state ) is
          when IDLE =>
            if matchFifo_empty = '0' then
              next_state <= MEM_INIT;
            else
              next_state <= IDLE;
            end if;
          when MEM_INIT => next_state <= WAIT_FINDER;
          when WAIT_FINDER =>
            if next_match_read_ptr >= matchFifo_write_ptr and finder_length_done = '1' then
              next_state <= IDLE;
            else
              -- todo jump to IDLE
              next_state <= WAIT_FINDER;
            end if;
          when SKIP_MATCHES =>

          when others =>
            next_state <= IDLE;
        end case; -- end case current_state
      end process; -- end MATCH_PROC_CTRL_KOMB_NEXT


      MATCH_PROC_CTRL_KOMB_OUT : process(all)
      begin

        case( current_state ) is
          when IDLE =>
            finder_list_match <= '0';
          when MEM_INIT =>
            finder_list_match <= '1';
          when WAIT_FINDER =>
            if matchFifo_empty = '0' and matchFifo_empty_reg = '0' then
              finder_list_match <= '1';
            end if;
          when others =>

        end case; -- end case current_state
      end process; -- end MATCH_PROC_CTRL_KOMB_OUT

      CONSECUTIVE_MATCHES : process(all)

        variable sI : natural;
        variable cI : natural;
      begin
        finder_consecutive_addresses <= (0 => '1', others => '0');
        for i in 1 to C_MATCH_COUNT-1-1 loop -- To avoid bug when we dont have match for next clock
          sI := HighestOne(finder_consecutive_addresses_reg)-1;
          cI := i+sI;
          if cI < C_MATCH_COUNT-1 and
             matchFifo_read_ptr_reg + cI < matchFifo_write_ptr and
             to_integer(unsigned(current_bus_match_source_addr((sI+1)*LZ4Address'length-1 downto sI*LZ4Address'length)))+i =
             to_integer(unsigned(current_bus_match_source_addr((cI+1)*LZ4Address'length-1 downto cI*LZ4Address'length))) and
             to_integer(unsigned(current_bus_match_addr((sI+1)*LZ4Address'length-1 downto sI*LZ4Address'length)))+i =
             to_integer(unsigned(current_bus_match_addr((cI+1)*LZ4Address'length-1 downto cI*LZ4Address'length)))
            then
            finder_consecutive_addresses(i) <= '1';
          else
            exit;
          end if;
        end loop;
      end process; -- end CONSECUTIVE_MATCHES
  end block MATCH_PROCESSOR_CONTROLLER;

  finder_ce <= out_ce;
  MatchLengthFinder_i : MatchLengthFinder
  generic map (
    MEMORY_BUS_WIDTH => C_MEMORY_BUS_WIDTH,
    MATCH_COUNT => C_MATCH_COUNT
  )
  port map (
    clk             => clk,
    rst             => rst,
    ce              => finder_ce,
    list_match      => finder_list_match,
    match_address   => current_match_addr,
    current_address => current_match_source_addr,
    consecutive_addresses => finder_consecutive_addresses,
    mf_address_1    => finder_mf_address_1,
    mf_address_2    => finder_mf_address_2,
    mf_end_stop     => finder_mf_end_stop,
    mf_start_stop   => control_processed_match_address,
    mf_data_1       => finder_mf_data_1,
    mf_data_2       => finder_mf_data_2,
    match_length    => finder_match_length,
    length_done     => finder_length_done,
    match_length_valid    => finder_match_length_valid,
    out_match_address     => finder_out_match_address,
    out_current_address  =>  finder_out_current_address
  );

  REG_FINDER_OUT : process(clk)
  begin
    if rising_edge(clk) then
      if rst = '1' then
        finder_match_length_off <= (others => '0');
        finder_match_length_valid_off <= (others => '0');
        finder_out_match_address_off <= (others => '0');
        finder_out_current_address_off <= (others => '0');
      elsif(out_ce = '1') then
        finder_match_length_off <= finder_match_length;
        finder_match_length_valid_off <= finder_match_length_valid;
        finder_out_match_address_off <= finder_out_match_address;
        finder_out_current_address_off <= finder_out_current_address;
      end if;
    end if;
  end process; -- end REG_FINDER_OUT

  MatchProcessorController_i : MatchProcessorController
  generic map (
    C_MATCH_COUNT      => C_MATCH_COUNT,
    C_MEMORY_BUS_WIDTH => C_MEMORY_BUS_WIDTH
  )
  port map (
    clk                 => clk,
    rst                 => rst,
    in_ce               => out_ce,
    out_ce              => out_ce and out_match_ready,
    match_done          => finder_done,
    memory_length       => memory_length,
    match_length        => finder_match_length_off,
    match_length_valid  => finder_match_length_valid_off,
    out_match_address   => finder_out_match_address_off,
    out_current_address => finder_out_current_address_off,
    processed_match_address => control_processed_match_address,
    out_literal_address => out_literal_address,
    out_litetal_data_valid => out_litetal_data_valid,
    out_literal_length => out_literal_length,
    out_match_length => out_match_length,
    out_match_offset => out_match_offset,
    out_match_valid => out_match_valid,
    last_serve => last_serve
  );


end architecture; -- end Behavioral
