library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;
use work.all;
use work.LZ4CompressorLib.all;
use ieee.math_real.all;

entity MatchProcessorController is
  generic(
    C_MATCH_COUNT : natural := 8;
    C_MEMORY_BUS_WIDTH : natural := 128
  );
  port(
    clk: in  std_logic;
    rst : in  std_logic;
    in_ce : in std_logic;
    out_ce : in std_logic;
    match_done : in std_logic;
    memory_length : in LZ4Address;
    match_length : in std_logic_vector(C_MATCH_COUNT*LZ4MatchLength'length-1 downto 0);
    match_length_valid : in std_logic_vector(C_MATCH_COUNT-1 downto 0);
    out_match_address : in LZ4Address;
    out_current_address : in LZ4Address;
    processed_match_address : out LZ4Address;
    last_serve : out std_logic;

    out_literal_address : out LZ4Address;
    out_litetal_data_valid : out std_logic_vector(C_MEMORY_BUS_WIDTH/8-1 downto 0);
    out_literal_length : out LZ4MatchLength;
    out_match_length : out LZ4MatchLength;
    out_match_offset : out LZ4Address;
    out_match_valid : out std_logic
  );
end entity; -- end MatchProcessorController

architecture arch of MatchProcessorController is
  component WordBram
  generic (
    UNALIGNED_WRITE : boolean := false;
    PORT_WORD_WIDTH : natural := 8;
    BYTE_WIDTH      : natural := 8;
    WORDS_COUNT     : natural := 2048;
    DEBUG_VIEW      : boolean := true
  );
  port (
    clk        : in  std_logic;
    rst        : in  std_logic;
    addr_in    : in  unsigned(integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT))))-1 downto 0);
    addr_out   : in  unsigned(integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT)))) -1 downto 0);
    data_in_we : in  std_logic_vector(PORT_WORD_WIDTH-1 downto 0);
    data_in    : in  std_logic_vector(PORT_WORD_WIDTH*BYTE_WIDTH-1 downto 0);
    data_out   : out std_logic_vector(PORT_WORD_WIDTH*BYTE_WIDTH-1 downto 0)
  );
  end component WordBram;

  component MaxReduction
  generic (
    INPUT_POWER : natural := 3;
    DATA_WIDTH  : natural := 16
  );
  port (
    data_in  : in  std_logic_vector(DATA_WIDTH * 2**INPUT_POWER-1 downto 0);
    data_out : out std_logic_vector(DATA_WIDTH-1 downto 0)
  );
  end component MaxReduction;

  component PPSPacker
  generic (
    PORTS      : natural := 8;
    DATA_WIDTH : natural := 64
  );
  port (
    clk         : in  std_logic;
    data_in_we  : in  std_logic_vector(PORTS-1 downto 0);
    data_in     : in  std_logic_vector(DATA_WIDTH*PORTS-1 downto 0);
    data_out    : out std_logic_vector(DATA_WIDTH*PORTS-1 downto 0);
    data_out_we : out std_logic_vector(PORTS-1 downto 0)
  );
  end component PPSPacker;

  component ShiftRegister
  generic (
    regDepth  : natural := 6;
    dataWidth : natural := 32
  );
  port (
    clk  : in  STD_LOGIC;
    rst  : in  STD_LOGIC;
    ce   : in  STD_LOGIC;
    din  : in  STD_LOGIC_VECTOR (dataWidth-1 downto 0);
    dout : out STD_LOGIC_VECTOR (dataWidth-1 downto 0)
  );
  end component ShiftRegister;


  constant NUMBER_OF_INPUT_WORDS  : integer := 512/4;
  constant ADDR_IN_WIDTH : natural := integer(ceil(log2(real(C_MATCH_COUNT*NUMBER_OF_INPUT_WORDS))));
  constant ADDR_OUT_WIDTH : natural := integer(ceil(log2(real(C_MATCH_COUNT*NUMBER_OF_INPUT_WORDS))));

  signal matchFifo_read_ptr  : unsigned(ADDR_IN_WIDTH-1 downto 0);
  signal matchFifo_read_ptr_reg  : unsigned(ADDR_IN_WIDTH-1 downto 0);
  signal matchFifo_write_ptr : unsigned(ADDR_OUT_WIDTH-1 downto 0);
  signal matchFifo_empty : std_logic;
  signal matchFifo_empty_reg : std_logic;
  signal matchFifo_done : std_logic;

  signal matchAddressDataIn : std_logic_vector(C_MATCH_COUNT*LZ4Address'length*2-1 downto 0);
  signal matchAddressDataIn_packed : std_logic_vector(C_MATCH_COUNT*LZ4Address'length*2-1 downto 0);

  signal max_match_address_in : std_logic_vector(C_MATCH_COUNT*LZ4Address'length-1 downto 0);
  signal current_max_match_address : LZ4Address;
  signal processed_match_address_reg : LZ4Address;

  signal match_length_valid_filtered : std_logic_vector(C_MATCH_COUNT-1 downto 0);

  signal match_length_valid_packed : std_logic_vector(C_MATCH_COUNT-1 downto 0);
  signal match_length_valid_packed_ce : std_logic_vector(C_MATCH_COUNT-1 downto 0);
  signal match_length_packed : std_logic_vector(C_MATCH_COUNT*LZ4MatchLength'length-1 downto 0);
begin
  processed_match_address <= processed_match_address_reg;

  matchFifo_empty <= '0' when matchFifo_read_ptr /= matchFifo_write_ptr else '1';
  matchFifo_done <= '1' when matchFifo_empty = '1' and match_done = '1' else '0';

  MAX_MATCH_ADDRESS_PROCESS : process(all)
    variable tmpAdd : LZ4Address;
  begin
    max_match_address_in <= (others => '0');
    for i in 0 to C_MATCH_COUNT-1 loop
      if match_length_valid(i) = '1' then
        tmpAdd := (others => '0');
        tmpAdd := resize(resize(unsigned(match_length((i+1)*LZ4MatchLength'length-1 downto i*LZ4MatchLength'length)), out_current_address'length) +
        unsigned(out_current_address), tmpAdd'length);
        max_match_address_in((i+1)*LZ4Address'length-1 downto i*LZ4Address'length) <= std_logic_vector(tmpAdd);
      end if;
    end loop;
  end process; -- end MAX_MATCH_ADDRESS_PROCESS

  MaxReduction_i : MaxReduction
  generic map (
    INPUT_POWER => integer(ceil(log2(real(C_MATCH_COUNT)))),
    DATA_WIDTH  => LZ4Address'length
  )
  port map (
    data_in  => max_match_address_in,
    unsigned(data_out) => current_max_match_address
  );

  INPUT_PROCESSOR : process(clk)
  begin
    if rising_edge(clk) then
      if rst = '1' then
        matchFifo_write_ptr <= (others => '0');
        processed_match_address_reg <= (others => '0');
      elsif(in_ce = '1') then
        if(or_reduce(match_length_valid_filtered) = '1') then
          matchFifo_write_ptr <= matchFifo_write_ptr + HighestOne(match_length_valid_packed);
        end if;

        processed_match_address_reg <= MAX(processed_match_address_reg, current_max_match_address);
      end if;
    end if;
  end process; -- end INPUT_PROCESSOR

  MATCH_FILTER : process(all)
  begin
    match_length_valid_filtered <= (others => '0');
    for i in 0 to C_MATCH_COUNT-1 loop
      if match_length_valid(i) = '1'
         and  --unsigned(match_length((i+1)*LZ4MatchLength'length-1 downto i*LZ4MatchLength'length)) +
         out_current_address+i > processed_match_address_reg then
        match_length_valid_filtered(i) <= '1';
        exit; -- Mozna by to chtelo vzit vic nez jednu v moment kdyz jsou obe validni a neprekrivaji se.
      end if;
    end loop;
  end process; -- end MATCH_FILTER

  PPSPacker_Match_Length : PPSPacker
  generic map (
    PORTS      => C_MATCH_COUNT,
    DATA_WIDTH => LZ4MatchLength'length
  )
  port map (
    clk         => clk,
    data_in_we  => match_length_valid_filtered,
    data_in     => match_length,
    data_out    => match_length_packed,
    data_out_we => match_length_valid_packed
  );


  ADDRESS_MULTI : process(all)
  begin
    matchAddressDataIn <= (others => '0');
    for i in 0 to C_MATCH_COUNT-1 loop
      matchAddressDataIn((i+1)*LZ4Address'length*2-1 downto i*LZ4Address'length*2) <= std_logic_vector(out_match_address+i) & std_logic_vector(out_current_address+i);
    end loop;
  end process; -- end ADDRESS_MULTI

  PPSPacker_Addresses : PPSPacker
  generic map (
    PORTS      => C_MATCH_COUNT,
    DATA_WIDTH => LZ4MatchLength'length*2
  )
  port map (
    clk         => clk,
    data_in_we  => match_length_valid_filtered,
    data_in     => matchAddressDataIn,
    data_out    => matchAddressDataIn_packed,
    data_out_we => open
  );


  LITERAL_OUTPUT_PROCESSOR : block
    type ControlStateType is (IDLE, LIT_SERVE, LAST_SERVE_PREP, DONE);
    constant ADDRESS_STEP : natural := C_MEMORY_BUS_WIDTH/8;

    signal current_state : ControlStateType;
    signal next_state : ControlStateType;

    signal cm_match_origin : LZ4Address;
    signal cm_match_source : LZ4Address;
    signal cm_match_length : LZ4MatchLength;
    signal cm_matchlen_tmp_buff : std_logic_vector(LZ4Address'length*8-1 downto 0);
    signal cm_match_tmp_buff : std_logic_vector(LZ4Address'length*2*8-1 downto 0);

    signal literal_address_out : LZ4Address;
    signal litetal_data_valid_out : std_logic_vector(C_MEMORY_BUS_WIDTH/8-1 downto 0);
    signal literal_length_out : LZ4MatchLength;
    signal match_length_out : LZ4MatchLength;
    signal match_offset_out : LZ4Address;
    signal match_valid_out : std_logic;

    signal literal_index : LZ4Address;
    signal moved_read_ptr : std_logic;
  begin

    LUP_SEQ : process(clk)
    begin
      if rising_edge(clk) then
        if rst = '1' then
          current_state <= IDLE;
          matchFifo_read_ptr <= (others => '0');
          match_offset_out <= (others => '0');
          match_length_out <= (others => '0');
          literal_address_out <= (others => '0');
          matchFifo_empty_reg <= '1';
          moved_read_ptr <= '0';
          last_serve <= '0';
        elsif(out_ce = '1') then
          match_valid_out <= '0';
          current_state <= next_state;
          matchFifo_empty_reg <= matchFifo_empty;
          matchFifo_read_ptr_reg <= matchFifo_read_ptr;

          case( current_state ) is
            when IDLE =>
              literal_length_out <= cm_match_origin - literal_address_out;
              match_offset_out <= cm_match_source;
              match_length_out <= cm_match_length;
              literal_index <= (others => '0');
              moved_read_ptr <= '0';
            when LAST_SERVE_PREP =>
              last_serve <= '1';
              literal_length_out <= memory_length - literal_address_out;
            when LIT_SERVE =>
              literal_index <= literal_index + ADDRESS_STEP;

              if literal_index + 2*ADDRESS_STEP > literal_length_out and literal_index + ADDRESS_STEP < literal_length_out then
                  matchFifo_read_ptr <= matchFifo_read_ptr+1;
                  moved_read_ptr <= '1';
              end if;
              if literal_index + ADDRESS_STEP > literal_length_out then
                if(last_serve = '0') then
                  literal_address_out <= literal_address_out + literal_length_out + match_length_out;
                  match_valid_out <= '1';
                else
                  literal_address_out <= literal_address_out + literal_length_out;
                end if;
                if(moved_read_ptr = '0') then
                  matchFifo_read_ptr <= matchFifo_read_ptr+1;
                end if;
              end if;
            when others =>
          end case; -- end case current_state
        end if;
      end if;
    end process; -- end LUP_SEQ

    LUP_NEXT : process(all)
    begin
      case( current_state ) is
        when IDLE =>
          if matchFifo_empty_reg = '0' and matchFifo_empty = '0'
              and matchFifo_read_ptr_reg = matchFifo_read_ptr then
            next_state <= LIT_SERVE;
          else
            if matchFifo_done = '1' then
              next_state <= LAST_SERVE_PREP;
            else
              next_state <= IDLE;
            end if;
          end if;
        when LIT_SERVE =>
          if literal_index + ADDRESS_STEP > literal_length_out then
            if last_serve = '1' then
              next_state <= DONE;
            else
              next_state <= IDLE;
            end if;
          else
            next_state <= LIT_SERVE;
          end if;
        when LAST_SERVE_PREP =>
          next_state <= LIT_SERVE;
        when others =>
      end case; -- end case current_state
    end process; -- end LUP_NEXT_SEQ

    LUP_OUTPUT : process(all)
    begin
      litetal_data_valid_out <= (others => '0');
      case( current_state ) is
        when LIT_SERVE =>
          for i in 0 to ADDRESS_STEP-1 loop
            if literal_index + i < literal_length_out then
              litetal_data_valid_out(i) <= '1';
            end if;
          end loop;
        when others =>
      end case; -- end case current_state
    end process; -- end LUP_OUTPUT



    OUTPUT_SHIFT_REG_BLOCK : block
      constant SHIFT_WIDTH : natural :=
        litetal_data_valid_out'length + literal_length_out'length +
        match_length_out'length + match_offset_out'length +
        1;
      signal shift_out : std_logic_vector(SHIFT_WIDTH-1 downto 0);
      signal slv_out_literal_length : std_logic_vector(literal_length_out'length-1 downto 0);
      signal slv_out_match_length : std_logic_vector(match_length_out'length-1 downto 0);
      signal slv_out_match_offset : std_logic_vector(match_offset_out'length-1 downto 0);
    begin
      ShiftRegister_i : ShiftRegister
      generic map (
        regDepth  => 2,
        dataWidth => SHIFT_WIDTH
      )
      port map (
        clk  => clk,
        rst  => rst,
        ce   => out_ce,
        din  => litetal_data_valid_out &
                std_logic_vector(literal_length_out) &
                std_logic_vector(match_length_out) &
                std_logic_vector(match_offset_out) &
                match_valid_out,
        dout => shift_out
      );
      process(all)
      begin
        if(out_ce = '1') then
          (out_litetal_data_valid,
           slv_out_literal_length,
           slv_out_match_length,
           slv_out_match_offset,
           out_match_valid) <= shift_out;
        else
         out_litetal_data_valid <= (others => '0');
         slv_out_literal_length <= (others => '0');
         slv_out_match_length <= (others => '0');
         slv_out_match_offset <= (others => '0');
         out_match_valid <= '0';
        end if;
      end process;

       out_literal_length <= unsigned(slv_out_literal_length) when out_ce = '1' else (others => '0');
       out_match_length   <= unsigned(slv_out_match_length) when out_ce = '1' else (others => '0');
       out_match_offset   <= unsigned(slv_out_match_offset) when out_ce = '1' else (others => '0');
    end block OUTPUT_SHIFT_REG_BLOCK;
    out_literal_address <= (literal_address_out + literal_index);

    match_length_valid_packed_ce <= match_length_valid_packed when in_ce = '1' else (others => '0');
    WordBram_MatchLengthMem : WordBram
    generic map (
      UNALIGNED_WRITE => true,
      PORT_WORD_WIDTH  => C_MATCH_COUNT,
      BYTE_WIDTH => LZ4Address'length,
      WORDS_COUNT => NUMBER_OF_INPUT_WORDS
    )
    port map (
      clk        => clk,
      rst        => rst,
      addr_in    => matchFifo_write_ptr,
      addr_out   => matchFifo_read_ptr,
      data_in_we => match_length_valid_packed_ce,
      data_in    => match_length_packed,
      data_out   => cm_matchlen_tmp_buff
    );
    cm_match_length <= unsigned(cm_matchlen_tmp_buff(LZ4Address'length-1 downto 0));

    WordBram_MatchAddrMem : WordBram
    generic map (
      UNALIGNED_WRITE => true,
      PORT_WORD_WIDTH  => C_MATCH_COUNT,
      BYTE_WIDTH => LZ4Address'length*2,
      WORDS_COUNT => NUMBER_OF_INPUT_WORDS
    )
    port map (
      clk        => clk,
      rst        => rst,
      addr_in    => matchFifo_write_ptr,
      addr_out   => matchFifo_read_ptr,
      data_in_we => match_length_valid_packed_ce,
      data_in    => matchAddressDataIn_packed,
      data_out   => cm_match_tmp_buff
    );

    cm_match_source <= unsigned(cm_match_tmp_buff(LZ4Address'length*2-1 downto LZ4Address'length));
    cm_match_origin <= unsigned(cm_match_tmp_buff(LZ4Address'length-1 downto 0));
  end block LITERAL_OUTPUT_PROCESSOR;
end architecture; -- end
