library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;
use work.all;
use std.textio.all;

entity WordBram is
  generic (
    UNALIGNED_WRITE : boolean := false;
    PORT_WORD_WIDTH : natural := 8;
    BYTE_WIDTH : natural := 8;
    WORDS_COUNT : natural := 2048;
    RAM_STYLE_STR : string := "block";
    DEBUG_VIEW : boolean := true;
    INIT_FILE : string := ""
  );
  port(
    clk : in std_logic;
    rst : in std_logic;
    addr_in : in  unsigned(integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT))))-1 downto 0);
    addr_out : in unsigned(integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT))))-1 downto 0);
    data_in_we : in std_logic_vector(PORT_WORD_WIDTH-1 downto 0);
    data_in : in std_logic_vector(PORT_WORD_WIDTH*BYTE_WIDTH-1 downto 0);
    data_out : out std_logic_vector(PORT_WORD_WIDTH*BYTE_WIDTH-1 downto 0)
  );
end entity; -- end WordBram

architecture Behavioral of WordBram is
  constant DATA_WIDTH : natural := PORT_WORD_WIDTH*BYTE_WIDTH;
  constant BYTE_COUNT : natural := PORT_WORD_WIDTH*WORDS_COUNT;
  constant ADDR_IN_WIDTH : natural := integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT))));
  constant ADDR_OUT_WIDTH : natural := integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT))));
  constant OFFSET_WIDTH : natural := integer(ceil(log2(real(PORT_WORD_WIDTH))));

  signal last_addr  : unsigned(ADDR_OUT_WIDTH-1 downto 0);
  signal barell_out_out : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal barell_out_in  : std_logic_vector(DATA_WIDTH-1 downto 0);

  signal barell_in_out : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal barell_in_in  : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal barell_we_in_out : std_logic_vector(PORT_WORD_WIDTH-1 downto 0);
  signal barell_we_in_in  : std_logic_vector(PORT_WORD_WIDTH-1 downto 0);

  type BYTE_FIFO_TYPE is array (0 to (BYTE_COUNT-1)) of std_logic_vector(BYTE_WIDTH-1 downto 0);
  attribute ram_style: string;

  signal offset_addr_in : unsigned(OFFSET_WIDTH-1 downto 0);
  signal offset_addr_out : unsigned(OFFSET_WIDTH-1 downto 0);
  signal offset_last_addr_out : unsigned(OFFSET_WIDTH-1 downto 0);


  type DEBUG_BYTE_MEM_TYPE is array(0 to (2**ADDR_OUT_WIDTH)-1) of std_logic_vector(BYTE_WIDTH-1 downto 0);
  signal debug_memory : DEBUG_BYTE_MEM_TYPE;
begin
  offset_addr_in <= addr_in(OFFSET_WIDTH-1 downto 0);
  offset_addr_out <= addr_out(OFFSET_WIDTH-1 downto 0);
  offset_last_addr_out <= last_addr(OFFSET_WIDTH-1 downto 0);


  BARELL_ADDR : process(clk)
  begin
    if rising_edge(clk) then
      if rst = '1' then
        last_addr <= (others => '0');
      else
        last_addr <= addr_out;
      end if;
    end if;
  end process; -- end BARELL_ADDR

  BARELL_SHIFT_IN_GEN : if UNALIGNED_WRITE = true generate
    barell_in_in <= data_in;
    barell_we_in_in <= data_in_we;

    BARELL_SHIFT_IN : process(barell_in_in, addr_in, offset_addr_in, barell_we_in_in)
      variable shift_dst : unsigned(OFFSET_WIDTH-1 downto 0);
    begin
      INDEX_LOOP : for i in 0 to PORT_WORD_WIDTH-1 loop
        shift_dst := to_unsigned(i, shift_dst'length) + offset_addr_in;
        barell_in_out((to_integer(shift_dst)+1)*BYTE_WIDTH-1 downto to_integer(shift_dst)*BYTE_WIDTH) <= barell_in_in((i+1)*BYTE_WIDTH-1 downto i*BYTE_WIDTH);

        barell_we_in_out(to_integer(shift_dst)) <= barell_we_in_in(i);
      end loop;  -- end for INDEX_LOOP
    end process; -- end BARELL_SHIFT_IN
  end generate; -- end if generate BARELL_SHIFT_IN_GEN

  BARELL_SHIFT_OUT : process(barell_out_in, last_addr, offset_last_addr_out)
    variable shift_dst : unsigned(OFFSET_WIDTH-1 downto 0);
  begin
    INDEX_LOOP : for i in 0 to PORT_WORD_WIDTH-1 loop
      shift_dst := to_unsigned(i, shift_dst'length) + to_unsigned((PORT_WORD_WIDTH-to_integer(offset_last_addr_out)), shift_dst'length);
      barell_out_out((to_integer(shift_dst)+1)*BYTE_WIDTH-1 downto to_integer(shift_dst)*BYTE_WIDTH) <= barell_out_in((i+1)*BYTE_WIDTH-1 downto i*BYTE_WIDTH);
    end loop;  -- end for INDEX_LOOP
  end process; -- end BARELL_SHIFT_IN

  BLOCK_RAM_GEN : for i in 0 to PORT_WORD_WIDTH-1 generate
    signal data_mem : BYTE_FIFO_TYPE;
    attribute ram_style of data_mem : signal is RAM_STYLE_STR;

    signal mem_addr_in : unsigned(ADDR_OUT_WIDTH-1 downto 0);
    signal mem_addr_out : unsigned(ADDR_OUT_WIDTH-1 downto 0);
    signal mem_data_in : std_logic_vector(BYTE_WIDTH-1 downto 0);
    signal mem_data_we_in : std_logic;
  begin
    MEM_OUT_ADDR : process(addr_out, offset_addr_out)
    begin
      mem_addr_out <= (others => '0');
      mem_addr_out(mem_addr_out'length-OFFSET_WIDTH-1 downto 0) <= addr_out(last_addr'length-1 downto OFFSET_WIDTH);
      if to_integer(offset_addr_out) > i then
        mem_addr_out(mem_addr_out'length-OFFSET_WIDTH-1 downto 0) <= addr_out(last_addr'length-1 downto OFFSET_WIDTH)+1;
      end if;
    end process; -- end MEM_OUT_ADDR

    MEM_UNALIGNED : if UNALIGNED_WRITE = true generate
      MEM_IN_ADDR : process(addr_in, offset_addr_in)
      begin
        mem_addr_in <= (others => '0');
        mem_addr_in(addr_in'length-OFFSET_WIDTH-1 downto 0) <= addr_in(addr_in'length-1 downto OFFSET_WIDTH);
        if to_integer(offset_addr_in) > i then
          mem_addr_in(addr_in'length-OFFSET_WIDTH-1 downto 0) <= addr_in(addr_in'length-1 downto OFFSET_WIDTH)+1;
        end if;
      end process; -- end MEM_ADDR
      mem_data_in <= barell_in_out((i+1)*BYTE_WIDTH-1 downto i*BYTE_WIDTH);
      mem_data_we_in <= barell_we_in_out(i);
    end generate; -- end if generate MEM_ADDR_IN_UNALIGNED

    MEM_ALIGNED : if UNALIGNED_WRITE = false generate
      mem_addr_in <= addr_in;
      mem_data_in <= data_in((i+1)*BYTE_WIDTH-1 downto i*BYTE_WIDTH);
      mem_data_we_in <= data_in_we(i);
    end generate; -- end if generate MEM_ADDR_IN_ALIGNED

    MEM_WRITE : process(clk)
      variable VEC_LINE : line;
      variable VEC_VAR : std_logic_vector(BYTE_WIDTH-1 downto 0);
      variable VEC_I : natural;
      variable VEC_M : natural;
      file VEC_FILE : text;
    begin
      if rising_edge(clk) then
        if rst = '1' then
          if(INIT_FILE /= "") then
            file_open(VEC_FILE, INIT_FILE, READ_MODE);
            VEC_I := 0;
            VEC_M := 0;
            while not endfile(VEC_FILE) loop
              readline (VEC_FILE, VEC_LINE);
              hread (VEC_LINE, VEC_VAR);
              if((VEC_I mod PORT_WORD_WIDTH) = i) then
                data_mem(VEC_M) <= VEC_VAR;
                VEC_M := VEC_M + 1;
              end if;
              VEC_I := VEC_I + 1;
            end loop;
            file_close(VEC_FILE);
          end if;
        else
          if(mem_data_we_in = '1') then
            data_mem(to_integer(mem_addr_in)) <= mem_data_in;
          end if;
        end if;
      end if;
    end process; -- end MEM_WRITE

    MEM_READ : process(clk)
    begin
      if rising_edge(clk) then
        barell_out_in((i+1)*BYTE_WIDTH-1 downto i*BYTE_WIDTH) <= data_mem(to_integer(mem_addr_out));
      end if;
    end process; -- end MEM_READ
  end generate; -- end for generate BLOCK_RAM_GEN

  data_out <= barell_out_out;


  -- DEBUG VIEW
  DEBUG_VIEW_GEN : if DEBUG_VIEW = true generate
    DEBUG_VIEW_WRITE : process(clk)
      variable VEC_I : natural;
      variable VEC_VAR : std_logic_vector(BYTE_WIDTH-1 downto 0);
      variable VEC_LINE : line;
      file VEC_FILE : text;
    begin
      if rising_edge(clk) then
        if rst = '1' then
          if(INIT_FILE /= "") then
            file_open(VEC_FILE, INIT_FILE, READ_MODE);
            VEC_I := 0;
            while not endfile(VEC_FILE) loop
              readline (VEC_FILE, VEC_LINE);
              hread (VEC_LINE, VEC_VAR);
              debug_memory(VEC_I) <= VEC_VAR;
              VEC_I := VEC_I + 1;
            end loop;
            file_close(VEC_FILE);
          end if;
        else
          for i in 0 to PORT_WORD_WIDTH-1 loop
            if data_in_we(i) = '1' then
              if UNALIGNED_WRITE = TRUE then
                debug_memory(to_integer(addr_in+i)) <= data_in((i+1)*BYTE_WIDTH-1 downto i*BYTE_WIDTH);
              else
                debug_memory(to_integer(addr_in*PORT_WORD_WIDTH+i)) <= data_in((i+1)*BYTE_WIDTH-1 downto i*BYTE_WIDTH);
              end if;
            end if;
          end loop;
        end if;
      end if;
    end process; -- end DEBUG_VIEW_WRITE
  end generate; -- end if generate DEBUG_VIEW_GEN
end architecture; -- end Behavioral
