library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;
use ieee.std_logic_misc.all;
use work.all;

entity WordBram is
  generic (
    UNALIGNED_WRITE : boolean := false;
    PORT_WORD_WIDTH : natural := 8;
    BYTE_WIDTH : natural := 8;
    WORDS_COUNT : natural := 2048;
    DEBUG_VIEW : boolean := true
  );
  port(
    clk_in : in std_logic;
    clk_out : in std_logic;
    rst_in : in std_logic;
    rst_out : in std_logic;
    addr_in : in  unsigned(integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT))))-1 downto 0);
    addr_out : in unsigned(integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT))))-1 downto 0);
    data_in_we : in std_logic_vector(PORT_WORD_WIDTH-1 downto 0);
    data_in : in std_logic_vector(PORT_WORD_WIDTH*BYTE_WIDTH-1 downto 0);
    data_out : out std_logic_vector(PORT_WORD_WIDTH*BYTE_WIDTH-1 downto 0);
    read_enable : in std_logic
  );
end entity; -- end WordBram

architecture Behavioral of WordBram is
  constant DATA_WIDTH : natural := 2**integer(ceil(log2(PORT_WORD_WIDTH*BYTE_WIDTH)));
  constant DATA_WIDTH_SIZE : natural := integer(ceil(log2(real(DATA_WIDTH))));
  constant ADDR_IN_WIDTH : natural := integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT))));
  constant ADDR_OUT_WIDTH : natural := integer(ceil(log2(real(PORT_WORD_WIDTH*WORDS_COUNT))));
  constant OFFSET_WIDTH : natural := integer(ceil(log2(real(PORT_WORD_WIDTH))));

  signal offset_addr_in : unsigned(OFFSET_WIDTH-1 downto 0);
  signal offset_addr_out : unsigned(OFFSET_WIDTH-1 downto 0);
  signal offset_last_addr_out : unsigned(OFFSET_WIDTH-1 downto 0);

  type DEBUG_BYTE_MEM_TYPE is array(0 to (2**ADDR_OUT_WIDTH)-1) of std_logic_vector(BYTE_WIDTH-1 downto 0);
  signal debug_memory : DEBUG_BYTE_MEM_TYPE;

  component ByteBram
  generic (
    C_RAM_WIDTH       : integer := 8;
    C_RAM_DEPTH       : integer := 512;
    C_RAM_PERFORMANCE : string := "LOW_LATENCY"
  );
  port (
    addra : in  std_logic_vector(clogb2(C_RAM_DEPTH)-1 downto 0);
    addrb : in  std_logic_vector(clogb2(C_RAM_DEPTH)-1 downto 0);
    dina  : in  std_logic_vector(C_RAM_WIDTH-1 downto 0);
    dinb  : in  std_logic_vector(C_RAM_WIDTH-1 downto 0);
    clka  : in  std_logic;
    clkb  : in  std_logic;
    wea   : in  std_logic;
    web   : in  std_logic;
    ena   : in  std_logic;
    enb   : in  std_logic;
    rsta  : in  std_logic;
    rstb  : in  std_logic;
    douta : out std_logic_vector(C_RAM_WIDTH-1 downto 0);
    doutb : out std_logic_vector(C_RAM_WIDTH-1 downto 0)
  );
  end component ByteBram;

  signal clk_read  : std_logic;
  signal clk_write  : std_logic;
  signal rst_read  : std_logic;
  signal rst_write  : std_logic;

  signal lower_addr_read : std_logic_vector(clogb2(C_RAM_DEPTH)-1 downto 0);
  signal lower_addr_write : std_logic_vector(clogb2(C_RAM_DEPTH)-1 downto 0);
  signal lower_din  : std_logic_vector(C_RAM_WIDTH-1 downto 0);
  signal lower_we   : std_logic;
  signal lower_re   : std_logic;
  signal lower_dout : std_logic_vector(C_RAM_WIDTH-1 downto 0);

  signal higher_addr_read : std_logic_vector(clogb2(C_RAM_DEPTH)-1 downto 0);
  signal higher_addr_write : std_logic_vector(clogb2(C_RAM_DEPTH)-1 downto 0);
  signal higher_din  : std_logic_vector(C_RAM_WIDTH-1 downto 0);
  signal higher_we   : std_logic;
  signal higher_re   : std_logic;
  signal higher_dout : std_logic_vector(C_RAM_WIDTH-1 downto 0);
begin
  offset_addr_in <= addr_in(OFFSET_WIDTH-1 downto 0);
  offset_addr_out <= addr_out(OFFSET_WIDTH-1 downto 0);
  offset_last_addr_out <= last_addr(OFFSET_WIDTH-1 downto 0);

  clk_read <= clk_out;
  clk_write <= clk_in;
  rst_read <= rst_out
  rst_write <= rst_in;

  LAST_ADDR_PROC : process(clk_out)
  begin
    if rising_edge(clk_out) then
      if rst_out = '1' then
        last_addr <= (others => '0');
      elsif(read_enable = '1') then
        last_addr <= addr_out;
      end if;
    end if;
  end process; -- end LAST_ADDR_PROC

  DATA_OUT_PROCESS : process(all)
  begin
    data_out <= lower_dout;
    if to_integer(offset_last_addr_out) /= 0 then
      data_out(data_out'length*BYTE_WIDTH-1 downto to_integer(offset_last_addr_out)*BYTE_WIDTH) <= lower_dout(lower_dout'length*BYTE_WIDTH-1 downto to_integer(offset_last_addr_out)*BYTE_WIDTH);
      data_out(to_integer(offset_last_addr_out)*BYTE_WIDTH-1 downto 0) <= higher_dout(to_integer(offset_last_addr_out)*BYTE_WIDTH-1 downto 0);
    end if;
  end process; -- end DATA_OUT_PROCESS

  lower_re <= read_enable;
  higher_re <= '0' when to_integer(offset_addr_out) = 0 else read_enable;
  lower_addr_read <= addr_out(addr_out'length-1 downto DATA_WIDTH_SIZE);
  higher_addr_read <= addr_out(addr_out'length-1 downto DATA_WIDTH_SIZE) + 1;

  DATA_IN_PROCESS : process(all)
  begin
    lower_we <= data_in_we;
    lower_din <= data_in;
    higher_we <= (others => '0');
    if to_integer(offset_addr_in) /= 0 then
      lower_we(lower_we'length-1 downto to_integer(offset_addr_out)) <= data_in_we(data_in_we'length-1 downto to_integer(offset_addr_out));
      higher_we(to_integer(offset_addr_out)-1 downto 0) <= data_in_we(to_integer(offset_addr_out)-1 downto 0);
      lower_din(lower_we'length*BYTE_WIDTH-1 downto to_integer(offset_addr_out)*BYTE_WIDTH) <= data_in(lower_we'length*BYTE_WIDTH-1 downto to_integer(offset_addr_out)*BYTE_WIDTH);
      higher_din(to_integer(offset_addr_out)*BYTE_WIDTH-1 downto 0) <= data_in(to_integer(offset_addr_out)*BYTE_WIDTH-1 downto 0);
    end if;
  end process; -- end DATA_IN_PROCESS

  lower_addr_write <= addr_in(addr_out'length-1 downto DATA_WIDTH_SIZE);
  higher_addr_write <= addr_in(addr_out'length-1 downto DATA_WIDTH_SIZE) + 1;
  ByteBram_i : ByteBram
  generic map (
    C_RAM_WIDTH       => DATA_WIDTH,
    C_RAM_DEPTH       => WORDS_COUNT,
    C_RAM_PERFORMANCE => "LOW_LATENCY"
  )
  port map (
    addra => lower_addr_read,
    addrb => lower_addr_write,
    dina  => (others => '0'),
    dinb  => lower_din,
    clka  => clk_read,
    clkb  => clk_write,
    wea   => (others => '0'),
    web   => lower_we,
    ena   => lower_re,
    enb   => '1',
    rsta  => rst_read,
    rstb  => rst_write,
    douta => lower_dout,
    doutb => open
  );

  ByteBram_i : ByteBram
  generic map (
    C_RAM_WIDTH       => DATA_WIDTH,
    C_RAM_DEPTH       => WORDS_COUNT,
    C_RAM_PERFORMANCE => "LOW_LATENCY"
  )
  port map (
    addra => higher_addr_read,
    addrb => higher_addr_write,
    dina  => (others => '0'),
    dinb  => higher_din,
    clka  => clk_read,
    clkb  => clk_write,
    wea   => (others => '0'),
    web   => higher_we,
    ena   => higher_re,
    enb   => '1',
    rsta  => rst_read,
    rstb  => rst_write,
    douta => higher_dout,
    doutb => open
  );

  -- DEBUG VIEW
  DEBUG_VIEW_GEN : if DEBUG_VIEW = true generate
    DEBUG_VIEW_WRITE : process(clk)
    begin
      if rising_edge(clk) then
        if rst /= '1' then
          for i in 0 to PORT_WORD_WIDTH-1 loop
            if data_in_we(i) = '1' then
              debug_memory(to_integer(addr_in+i)) <= data_in((i+1)*BYTE_WIDTH-1 downto i*BYTE_WIDTH);
            end if;
          end loop;
        end if;
      end if;
    end process; -- end DEBUG_VIEW_WRITE
  end generate; -- end if generate DEBUG_VIEW_GEN
end architecture; -- end Behavioral
